"""TcEx Framework Module"""

# first-party
from tcex.api.tc.ti_transform.model.transform_model import (
    AttributeTransformModel,
    DatetimeTransformModel,
    GroupTransformModel,
    IndicatorTransformModel,
    MetadataTransformModel,
    SecurityLabelTransformModel,
    TagTransformModel,
    TiTransformModel,
    TransformModel,
)

__all__ = [
    'AttributeTransformModel',
    'DatetimeTransformModel',
    'GroupTransformModel',
    'IndicatorTransformModel',
    'MetadataTransformModel',
    'SecurityLabelTransformModel',
    'TagTransformModel',
    'TiTransformModel',
    'TransformModel',
]



"""TcEx Framework Module"""
# standard library
import logging
import os
import sys
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Literal, NoReturn

# first-party
from tcex.app.config import InstallJson
from tcex.logger.trace_logger import TraceLogger
from tcex.registry import registry

if TYPE_CHECKING:
    # first-party
    from tcex.input.input import Input  # CIRCULAR-IMPORT


# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


EXIT_CODES = Literal[0, 1, 3, 4]


class ExitCode(int, Enum):
    """Valid exit codes for a ThreatConnect app.

    Note not all exit codes are valid for all app types:
    partial failure is not valid for playbook apps.
    """

    SUCCESS = 0
    FAILURE = 1
    PARTIAL_FAILURE = 3
    HARD_FAILURE = 4

    def __str__(self):
        """@cblades"""
        clean_name = self.name.replace('_', ' ').title()
        return f'{self.value} ({clean_name})'


class Exit:
    """Provides functionality around exiting an app."""

    def __init__(self, inputs: 'Input'):
        """Initialize instance properties."""
        self.ij = InstallJson()
        self.inputs = inputs

        # properties
        self._exit_code = ExitCode.SUCCESS
        self._message = None
        self.log = _logger

    def _aot_rpush(self, exit_code: int):
        """Push message to AOT action channel."""
        if self.inputs.contents.get('tc_playbook_db_type') == 'Redis':
            try:
                # pylint: disable=no-member
                registry.redis_client.rpush(self.inputs.contents['tc_exit_channel'], exit_code)
            except Exception as e:  # pragma: no cover
                self._exit(ExitCode.FAILURE, f'Exception during AOT exit push ({e}).')

    def _exit(self, code: ExitCode | int, msg: str) -> NoReturn:
        """Exit the App"""
        code = ExitCode(code) if code is not None else self.code

        # handle exit msg logging
        self._exit_msg_handler(code, msg)

        self.log.info(f'exit-code={code}')
        sys.exit(code.value)

    def _message_tc(self, message: str, max_length: int = 255):
        """Write data to message_tc file in TcEX specified directory.

        This method is used to set and exit message in the ThreatConnect Platform.
        ThreatConnect only supports files of max_message_length.  Any data exceeding
        this limit will be truncated. The last <max_length> characters will be preserved.

        Args:
            message: The message to add to message_tc file
            max_length: The maximum length of an exit message. Defaults to 255.
        """
        if not isinstance(message, str):
            message = str(message)

        if os.access(self.inputs.model_tc.tc_out_path, os.W_OK):
            message_file = self.inputs.model_tc.tc_out_path / 'message.tc'
        else:
            message_file = Path('message.tc')

        if not message.endswith('\n'):
            message += '\n'
        with message_file.open('w') as mh:
            # write last <max_length> characters to file
            mh.write(message[-max_length:])

    @property
    def code(self) -> ExitCode:
        """Get exit code."""
        return self._exit_code

    @code.setter
    def code(self, exit_code: ExitCode | int):
        """Set exit code.

        Will automatically change partial failure to success if app is a playbook app.

        Args:
            exit_code: the new exit code.
        """
        exit_code = ExitCode(exit_code)
        if exit_code == ExitCode.PARTIAL_FAILURE and self.ij.model.is_playbook_app:
            self.log.info(
                f'Changing exit code from {ExitCode.PARTIAL_FAILURE} '
                f'to {ExitCode.SUCCESS} for Playbook App.'
            )
            exit_code = ExitCode.SUCCESS
        self._exit_code = exit_code

    def exit(self, code: ExitCode | int | None = None, msg: str | None = None) -> NoReturn:
        """Application exit method with proper exit code

        The method will run the Python standard sys.exit() with the exit code
        previously defined via :py:meth:`~tcex.tcex.TcEx.exit_code` or provided
        during the call of this method.

        Args:
            code: The exit code value for the app.
            msg: A message to log and add to message tc output.
        """
        code = ExitCode(code) if code is not None else self.code
        msg = msg if msg is not None else ''

        # playbook exit handler
        if self.ij.model.is_playbook_app:
            self.exit_playbook_handler(msg)

        # aot notify
        if 'tc_aot_enabled' in self.inputs.contents and self.inputs.contents.get('tc_aot_enabled'):
            # push exit message
            self._aot_rpush(code.value)

        # exit token renewal thread
        if not self.ij.is_external_app:
            registry.token.shutdown = True

        # exit
        self._exit(code, msg)

    # TODO: [med] @cblades - is msg required?
    # pylint: disable=unused-argument
    def exit_aot_terminate(self, code: ExitCode | int | None = None, msg: str | None = None):
        """Application exit method with proper exit code only for AOT.

        The method will run the Python standard sys.exit() with the exit code
        previously defined via :py:meth:`~tcex.tcex.TcEx.exit_code` or provided
        during the call of this method.

        Args:
            code: The exit code value for the app.
            msg: A message to log and add to message tc output.
        """
        code = ExitCode(code) if code is not None else self.code
        msg = msg if msg is not None else self.message

        # aot notify
        if 'tc_aot_enabled' in self.inputs.contents and self.inputs.contents.get('tc_aot_enabled'):
            # push exit message
            self._aot_rpush(code.value)

        self.log.info(f'exit-code={code}')
        sys.exit(code)

    def _exit_msg_handler(self, code: ExitCode, msg: str):
        """Handle exit message. Write to both log and message_tc."""
        if msg is not None:
            log_msg = msg.replace('\n', ',')
            if code in [ExitCode.SUCCESS, ExitCode.PARTIAL_FAILURE]:
                self.log.info(f'exit-message="{log_msg}"')
            else:
                self.log.error(f'exit-message="{log_msg}"')
            self._message_tc(msg)

    def exit_playbook_handler(self, msg: str):
        """Perform special action for PB Apps before exit."""
        # write outputs before exiting
        registry.playbook.output.process()  # pylint: disable=no-member

        # required only for tcex testing framework
        if (
            hasattr(self.inputs.model_tc, 'tcex_testing_context')
            and self.inputs.model_tc.tcex_testing_context is not None
        ):  # pragma: no cover
            registry.redis_client.hset(  # pylint: disable=no-member
                self.inputs.model_tc.tcex_testing_context, '_exit_message', msg
            )

    @property
    def message(self) -> str:
        """Get exit code."""
        return self._message or ''

    @message.setter
    def message(self, message: str):
        """Set exit message."""
        self._message = message



"""Model Definition"""
# standard library
from collections.abc import Callable

# third-party
from jmespath import compile as jmespath_compile
from pydantic import BaseModel, Extra, Field, root_validator, validator


# reusable validator
def _always_array(value: list | str) -> list[str]:
    """Return value as a list."""
    if not isinstance(value, list):
        value = [value]
    return value


# pylint: disable=no-self-argument
class TransformModel(BaseModel, extra=Extra.forbid):
    """Model Definition"""

    filter_map: dict | None = Field(None, description='')
    kwargs: dict | None = Field({}, description='')
    method: Callable | None = Field(None, description='')
    for_each: Callable | None = Field(None, description='')
    static_map: dict | None = Field(None, description='')

    @validator('filter_map', 'static_map', pre=True)
    def _lower_map_keys(cls, v):
        """Validate static_map."""
        if isinstance(v, dict):
            v = {key.lower(): value for (key, value) in v.items()}
        return v

    # root validator that ensure at least one indicator value/summary is set
    @root_validator()
    def _required_transform(cls, v: dict):
        """Validate either static_map or method is defined."""

        fields = [v.get('filter_map'), v.get('static_map'), v.get('method'), v.get('for_each')]

        if not any(fields):
            raise ValueError('Either map or method must be defined.')

        if fields.count(None) < 2:
            raise ValueError('Only one transform mechanism can be defined.')
        return v


class PathTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    default: str | None = Field(None, description='')
    path: str | None = Field(None, description='')

    @validator('path')
    def _validate_path(cls, v):
        """Validate path."""
        if v is not None:
            try:
                _ = jmespath_compile(v)
            except Exception:
                raise ValueError('A valid path must be provided.')
        return v

    # root validator that ensure at least one indicator value/summary is set
    @root_validator()
    def _default_or_path(cls, values):
        """Validate either default or path is defined."""
        if not any([values.get('default'), values.get('path')]):
            raise ValueError('Either default or path or both must be defined.')
        return values


class MetadataTransformModel(PathTransformModel, extra=Extra.forbid):
    """."""

    transform: list[TransformModel] | None = Field(None, description='')

    # validators
    _transform_array = validator('transform', allow_reuse=True, pre=True)(_always_array)


class ValueTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    value: str | MetadataTransformModel


class FileOccurrenceTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    file_name: str | MetadataTransformModel | None = Field(None, description='')
    path: str | MetadataTransformModel | None = Field(None, description='')
    date: str | MetadataTransformModel | None = Field(None, description='')


class DatetimeTransformModel(PathTransformModel, extra=Extra.forbid):
    """."""


class AssociatedGroupTransform(ValueTransformModel, extra=Extra.forbid):
    """."""


class AttributeTransformModel(ValueTransformModel, extra=Extra.forbid):
    """."""

    displayed: bool | MetadataTransformModel = Field(False, description='')
    source: str | MetadataTransformModel | None = Field(None, description='')
    type: str | MetadataTransformModel = Field(..., description='')


class SecurityLabelTransformModel(ValueTransformModel, extra=Extra.forbid):
    """."""

    color: MetadataTransformModel | None = Field(None, description='')
    description: MetadataTransformModel | None = Field(None, description='')


class TagTransformModel(ValueTransformModel, extra=Extra.forbid):
    """."""


class TiTransformModel(BaseModel, extra=Extra.forbid):
    """."""

    applies: Callable | None = Field(None, description='')
    associated_groups: list[AssociatedGroupTransform] = Field([], description='')
    attributes: list[AttributeTransformModel] = Field([], description='')
    date_added: DatetimeTransformModel | None = Field(None, description='')
    last_modified: DatetimeTransformModel | None = Field(None, description='')
    security_labels: list[SecurityLabelTransformModel] = Field([], description='')
    tags: list[TagTransformModel] = Field([], description='')
    type: MetadataTransformModel = Field(..., description='')
    xid: MetadataTransformModel | None = Field(None, description='')


class GroupTransformModel(TiTransformModel, extra=Extra.forbid):
    """."""

    name: MetadataTransformModel = Field(..., description='')
    # campaign
    first_seen: DatetimeTransformModel | None = Field(None, description='')
    # document
    malware: MetadataTransformModel | None = Field(None, description='')
    password: MetadataTransformModel | None = Field(None, description='')
    # email
    from_addr: MetadataTransformModel | None = Field(None, description='')
    score: MetadataTransformModel | None = Field(None, description='')
    to_addr: MetadataTransformModel | None = Field(None, description='')
    subject: MetadataTransformModel | None = Field(None, description='')
    body: MetadataTransformModel | None = Field(None, description='')
    header: MetadataTransformModel | None = Field(None, description='')
    # event, incident
    event_date: DatetimeTransformModel | None = Field(None, description='')
    status: MetadataTransformModel | None = Field(None, description='')
    # report
    publish_date: DatetimeTransformModel | None = Field(None, description='')
    # signature
    file_type: MetadataTransformModel | None = Field(None, description='')
    file_text: MetadataTransformModel | None = Field(None, description='')
    # document, signature
    file_name: MetadataTransformModel | None = Field(None, description='')


# pylint: disable=no-self-argument
class IndicatorTransformModel(TiTransformModel, extra=Extra.forbid):
    """."""

    confidence: MetadataTransformModel | None = Field(None, description='')
    rating: MetadataTransformModel | None = Field(None, description='')
    # summary: MetadataTransformModel | None = Field(None, description='')
    value1: MetadataTransformModel | None = Field(None, description='')
    value2: MetadataTransformModel | None = Field(None, description='')
    value3: MetadataTransformModel | None = Field(None, description='')
    # file
    file_occurrences: list[FileOccurrenceTransformModel] | None = Field(None, description='')
    size: MetadataTransformModel | None = Field(None, description='')
    # host
    dns_active: MetadataTransformModel | None = Field(None, description='')
    whois_active: MetadataTransformModel | None = Field(None, description='')
    active: MetadataTransformModel | None = Field(None, description='')

    # root validator that ensure at least one indicator value/summary is set
    @root_validator()
    def _one_indicator_value(cls, values):
        """Validate that one set of credentials is provided for the TC API."""

        if not any(
            [
                values.get('value1'),
                values.get('value2'),
                values.get('value3'),
            ]
        ):
            raise ValueError(
                'An indicator transform must be defined (e.g., summary, value1, value2, value3).'
            )
        return values



"""TcEx Framework Module"""
# standard library
import re
from urllib.parse import urlsplit

# third-party
from pydantic import BaseModel, Field
from requests import PreparedRequest  # TYPE-CHECKING

from .util import Util


class CurlModel(BaseModel):
    """Model Definition"""

    body_limit: int = Field(100, description='The size limit for the body value.')
    mask_body: bool = Field(False, description='If True the body will be masked.')
    proxies: dict[str, str] | None = Field(
        None, description='A dict containing the proxy configuration.'
    )
    verify: bool | str = Field(
        True, description='If False the curl command will include --insecure flag.'
    )
    write_file: bool = Field(
        False,
        description='If True and the body is binary it will be written as a temp file.',
    )


class RequestsToCurl:
    """TcEx Utilities Request to Curl Class"""

    def __init__(self):
        """Initialize the Class properties."""
        self.util = Util()

    def _process_body(self, body: bytes | str | None, curl_model: CurlModel) -> list[str]:
        """Process body and return a list of curl commands."""
        _body = []
        if body is None:
            return _body

        # add body to the curl command
        if isinstance(body, str):
            if curl_model.mask_body is True:
                body = self.util.printable_cred(body)  # mask the body
            else:
                body = self.util.truncate_string(
                    string=body, length=curl_model.body_limit, append_chars='...'
                )
            _body.append(f'-d "{body}"')
        elif isinstance(body, bytes):
            _body.append('--data-binary @/tmp/body-file')
        return _body

    def _process_headers(
        self, headers: dict, mask_headers: bool, mask_patterns: list[str] | None
    ) -> list[str]:
        """Process headers and return a list of curl commands."""
        _headers = []
        for k, v in sorted(dict(headers).items()):
            if mask_headers is True:
                patterns = [
                    'authorization',
                    'cookie',
                    'password',
                    'secret',
                    'session',
                    'token',
                    'username',
                ]
                if isinstance(mask_patterns, list):
                    # add user defined mask patterns
                    patterns.extend(mask_patterns)

                for p in patterns:
                    if re.match(rf'.*{p}.*', k, re.IGNORECASE):
                        v: str = self.util.printable_cred(v)

                # using gzip in Accept-Encoding with CURL on the CLI produces
                # the warning "Binary output can mess up your terminal."
                if k.lower() == 'accept-encoding':
                    encodings = [e.strip() for e in v.split(',')]
                    for encoding in list(encodings):
                        if encoding in ['gzip']:
                            encodings.remove(encoding)
                    v: str = ', '.join(encodings)

            _headers.append(f"-H '{k}: {v}'")
        return _headers

    def _process_proxies(self, proxies: dict[str, str] | None) -> list[str]:
        """Process proxies and return a list of curl commands."""
        _proxies = []
        if proxies is None or proxies.get('https') is None:
            return _proxies

        # parse formatted string {'https': 'bob:pass@https://localhost:4242'}
        proxy_url = proxies.get('https')
        proxy_data = urlsplit(proxy_url)

        # server
        proxy_server = proxy_data.hostname
        # do not process if server is None or bytes
        if isinstance(proxy_server, str):
            if proxy_data.port:
                proxy_server = f'{proxy_data.hostname}:{proxy_data.port}'
            _proxies.extend(['--proxy', proxy_server])

            # only append auth if server name is str
            if proxy_data.username:
                _proxies.extend(['--proxy-user', f'{proxy_data.username}:xxxxx'])

        return _proxies

    def convert(
        self,
        request: PreparedRequest,
        mask_headers: bool = True,
        mask_patterns: list[str] | None = None,
        **kwargs,
    ) -> str:
        """Return converted Prepared Request to a curl command.

        Args:
            request: The response.request object.
            mask_headers: If True then values for certain header key will be masked.
            mask_patterns: A list of patterns if found in headers the value will be masked.
            **kwargs: Additional keyword arguments to pass to the CurlModel.

        Keyword Args:
            body_limit: The size limit for the body value.
            mask_body: If True the body will be masked.
            proxies: A dict containing the proxy configuration.
            verify: If False the curl command will include --insecure flag.
            write_file: If True and the body is binary it will be written as a temp file.
        """
        # build curl model from kwargs
        curl_model = CurlModel(**kwargs)

        # APP-79 - adding the ability to log request as curl commands
        cmd = ['curl', '-X', request.method]

        # add headers to curl command
        cmd.extend(self._process_headers(dict(request.headers), mask_headers, mask_patterns))

        # add body to the curl command
        cmd.extend(self._process_body(request.body, curl_model))

        # add proxies to curl command
        cmd.extend(self._process_proxies(curl_model.proxies))

        if curl_model.verify is not True:
            # add insecure flag to curl command
            cmd.append('--insecure')

        # add url to curl command
        cmd.append(request.url)

        return ' '.join(cmd)



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class VictimModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Victim Model',
    validate_assignment=True,
):
    """Victim Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    assets: 'VictimAssetsModel' = Field(
        None,
        description='A list of victim assets corresponding to the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='assets',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of groups that this victim is associated with.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    attributes: 'VictimAttributesModel' = Field(
        None,
        description='A list of Attributes corresponding to the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='attributes',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the item was first created.',
        read_only=True,
        title='dateAdded',
    )
    description: str | None = Field(
        None,
        allow_mutation=False,
        description='Description of the Victim.',
        read_only=True,
        title='description',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        description='Name of the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    nationality: str | None = Field(
        None,
        description='Nationality of the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='nationality',
    )
    org: str | None = Field(
        None,
        description='Org of the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='org',
    )
    owner_name: str | None = Field(
        None,
        allow_mutation=False,
        description='The name of the Organization, Community, or Source that the item belongs to.',
        read_only=True,
        title='ownerName',
    )
    security_labels: 'SecurityLabelsModel' = Field(
        None,
        description=(
            'A list of Security Labels corresponding to the Intel item (NOTE: Setting this '
            'parameter will replace any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='securityLabels',
    )
    suborg: str | None = Field(
        None,
        description='Suborg of the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='suborg',
    )
    tags: 'TagsModel' = Field(
        None,
        description=(
            'A list of Tags corresponding to the item (NOTE: Setting this parameter will replace '
            'any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='tags',
    )
    web_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the ThreatConnect details page for this entity.',
        read_only=True,
        title='webLink',
    )
    work_location: str | None = Field(
        None,
        description='Work location of the Victim.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='workLocation',
    )

    @validator('associated_groups', always=True, pre=True)
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('security_labels', always=True, pre=True)
    def _validate_security_labels(cls, v):
        if not v:
            return SecurityLabelsModel()  # type: ignore
        return v

    @validator('tags', always=True, pre=True)
    def _validate_tags(cls, v):
        if not v:
            return TagsModel()  # type: ignore
        return v

    @validator('assets', always=True, pre=True)
    def _validate_victim_assets(cls, v):
        if not v:
            return VictimAssetsModel()  # type: ignore
        return v

    @validator('attributes', always=True, pre=True)
    def _validate_victim_attributes(cls, v):
        if not v:
            return VictimAttributesModel()  # type: ignore
        return v


class VictimDataModel(
    BaseModel,
    title='Victim Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Victims Data Model"""

    data: list[VictimModel] | None = Field(
        [],
        description='The data for the Victims.',
        methods=['POST', 'PUT'],
        title='data',
    )


class VictimsModel(
    BaseModel,
    title='Victims Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Victims Model"""

    _mode_support = PrivateAttr(False)

    data: list[VictimModel] | None = Field(
        [],
        description='The data for the Victims.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelsModel
from tcex.api.tc.v3.tags.tag_model import TagsModel
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetsModel
from tcex.api.tc.v3.victim_attributes.victim_attribute_model import VictimAttributesModel

# add forward references
VictimDataModel.update_forward_refs()
VictimModel.update_forward_refs()
VictimsModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class SecurityLabelFilter(FilterABC):
    """Filter Object for SecurityLabels"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.SECURITY_LABELS.value

    def color(self, operator: Enum, color: list | str):
        """Filter Color based on **color** keyword.

        Args:
            operator: The operator enum for the filter.
            color: The color of the security label (in hex triplet format).
        """
        if isinstance(color, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('color', operator, color, TqlType.STRING)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the security label was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the security label.
        """
        if isinstance(description, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    def has_group_attribute(self, operator: Enum, has_group_attribute: int | list):
        """Filter Associated Group based on **hasGroupAttribute** keyword.

        Args:
            operator: The operator enum for the filter.
            has_group_attribute: A nested query for association to other groups.
        """
        if isinstance(has_group_attribute, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('hasGroupAttribute', operator, has_group_attribute, TqlType.INTEGER)

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator_filter import IndicatorFilter

        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    def has_indicator_attribute(self, operator: Enum, has_indicator_attribute: int | list):
        """Filter Associated Indicator based on **hasIndicatorAttribute** keyword.

        Args:
            operator: The operator enum for the filter.
            has_indicator_attribute: A nested query for association to other indicators.
        """
        if isinstance(has_indicator_attribute, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter(
            'hasIndicatorAttribute', operator, has_indicator_attribute, TqlType.INTEGER
        )

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victims.victim_filter import VictimFilter

        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    def has_victim_attribute(self, operator: Enum, has_victim_attribute: int | list):
        """Filter Associated Victim based on **hasVictimAttribute** keyword.

        Args:
            operator: The operator enum for the filter.
            has_victim_attribute: A nested query for association to other victims.
        """
        if isinstance(has_victim_attribute, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('hasVictimAttribute', operator, has_victim_attribute, TqlType.INTEGER)

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the security label.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the security label.
        """
        if isinstance(name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The owner ID of the security label.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the security label.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The name of the security label.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def victim_id(self, operator: Enum, victim_id: int | list):
        """Filter Victim ID based on **victimId** keyword.

        Args:
            operator: The operator enum for the filter.
            victim_id: The ID of the victim the security label is applied to.
        """
        if isinstance(victim_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('victimId', operator, victim_id, TqlType.INTEGER)



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class IndicatorFilter(FilterABC):
    """Filter Object for Indicators"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.INDICATORS.value

    def active_locked(self, operator: Enum, active_locked: bool):
        """Filter Indicator Status Locked based on **activeLocked** keyword.

        Args:
            operator: The operator enum for the filter.
            active_locked: Whether or not the indicator status is locked.
        """
        self._tql.add_filter('activeLocked', operator, active_locked, TqlType.BOOLEAN)

    def address_asn(self, operator: Enum, address_asn: int | list):
        """Filter ASN (Address) based on **addressAsn** keyword.

        Args:
            operator: The operator enum for the filter.
            address_asn: The Autonomous System Number (ASN) of an address.
        """
        if isinstance(address_asn, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressAsn', operator, address_asn, TqlType.INTEGER)

    def address_cidr(self, operator: Enum, address_cidr: list | str):
        """Filter CIDR (Address) based on **addressCidr** keyword.

        Args:
            operator: The operator enum for the filter.
            address_cidr: A CIDR block used to search for a range of addresses.
        """
        if isinstance(address_cidr, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressCidr', operator, address_cidr, TqlType.STRING)

    def address_city(self, operator: Enum, address_city: list | str):
        """Filter City (Address) based on **addressCity** keyword.

        Args:
            operator: The operator enum for the filter.
            address_city: The name of the city an address is registered to.
        """
        if isinstance(address_city, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressCity', operator, address_city, TqlType.STRING)

    def address_country_code(self, operator: Enum, address_country_code: list | str):
        """Filter Country Code (Address) based on **addressCountryCode** keyword.

        Args:
            operator: The operator enum for the filter.
            address_country_code: The registered country code for an address.
        """
        if isinstance(address_country_code, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressCountryCode', operator, address_country_code, TqlType.STRING)

    def address_country_name(self, operator: Enum, address_country_name: list | str):
        """Filter Country Name (Address) based on **addressCountryName** keyword.

        Args:
            operator: The operator enum for the filter.
            address_country_name: The name of the country an address is registered to.
        """
        if isinstance(address_country_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressCountryName', operator, address_country_name, TqlType.STRING)

    def address_ip_val(self, operator: Enum, address_ip_val: int | list):
        """Filter Value (Address) based on **addressIpVal** keyword.

        Args:
            operator: The operator enum for the filter.
            address_ip_val: The numeric value of an address.
        """
        if isinstance(address_ip_val, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressIpVal', operator, address_ip_val, TqlType.INTEGER)

    def address_is_ipv6(self, operator: Enum, address_is_ipv6: bool):
        """Filter Type (Address) based on **addressIsIpv6** keyword.

        Args:
            operator: The operator enum for the filter.
            address_is_ipv6: A boolean indicating if the address is of type Ipv6.
        """
        self._tql.add_filter('addressIsIpv6', operator, address_is_ipv6, TqlType.BOOLEAN)

    def address_registering_org(self, operator: Enum, address_registering_org: list | str):
        """Filter Registering Org (Address) based on **addressRegisteringOrg** keyword.

        Args:
            operator: The operator enum for the filter.
            address_registering_org: The registering organization for an address.
        """
        if isinstance(address_registering_org, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter(
            'addressRegisteringOrg', operator, address_registering_org, TqlType.STRING
        )

    def address_state(self, operator: Enum, address_state: list | str):
        """Filter State (Address) based on **addressState** keyword.

        Args:
            operator: The operator enum for the filter.
            address_state: The name of the state an address is registered to.
        """
        if isinstance(address_state, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressState', operator, address_state, TqlType.STRING)

    def address_timezone(self, operator: Enum, address_timezone: list | str):
        """Filter Time Zone (Address) based on **addressTimezone** keyword.

        Args:
            operator: The operator enum for the filter.
            address_timezone: The time zone an address resides within.
        """
        if isinstance(address_timezone, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('addressTimezone', operator, address_timezone, TqlType.STRING)

    def associated_group(self, operator: Enum, associated_group: int | list):
        """Filter associatedGroup based on **associatedGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            associated_group: No description provided.
        """
        if isinstance(associated_group, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('associatedGroup', operator, associated_group, TqlType.INTEGER)

    def attribute(self, operator: Enum, attribute: list | str):
        """Filter attribute based on **attribute** keyword.

        Args:
            operator: The operator enum for the filter.
            attribute: No description provided.
        """
        if isinstance(attribute, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('attribute', operator, attribute, TqlType.STRING)

    def confidence(self, operator: Enum, confidence: int | list):
        """Filter Confidence Rating based on **confidence** keyword.

        Args:
            operator: The operator enum for the filter.
            confidence: The confidence in the indicator's rating.
        """
        if isinstance(confidence, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('confidence', operator, confidence, TqlType.INTEGER)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the indicator was added to the system.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The default description of the indicator.
        """
        if isinstance(description, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def false_positive_count(self, operator: Enum, false_positive_count: int | list):
        """Filter False Positive Count based on **falsePositiveCount** keyword.

        Args:
            operator: The operator enum for the filter.
            false_positive_count: The number of times the indicator has been flagged as a false
                positive.
        """
        if isinstance(false_positive_count, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('falsePositiveCount', operator, false_positive_count, TqlType.INTEGER)

    def file_size(self, operator: Enum, file_size: int | list):
        """Filter Size (File) based on **fileSize** keyword.

        Args:
            operator: The operator enum for the filter.
            file_size: The size of a file.
        """
        if isinstance(file_size, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('fileSize', operator, file_size, TqlType.INTEGER)

    @property
    def has_artifact(self):
        """Return **ArtifactFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter

        artifacts = ArtifactFilter(Tql())
        self._tql.add_filter('hasArtifact', TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)
        return artifacts

    @property
    def has_attribute(self):
        """Return **IndicatorAttributeFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.indicator_attributes.indicator_attribute_filter import (
            IndicatorAttributeFilter,
        )

        attributes = IndicatorAttributeFilter(Tql())
        self._tql.add_filter('hasAttribute', TqlOperator.EQ, attributes, TqlType.SUB_QUERY)
        return attributes

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_group(self):
        """Return **GroupFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.groups.group_filter import GroupFilter

        groups = GroupFilter(Tql())
        self._tql.add_filter('hasGroup', TqlOperator.EQ, groups, TqlType.SUB_QUERY)
        return groups

    @property
    def has_indicator(self):
        """Return **IndicatorFilter** for further filtering."""
        indicators = IndicatorFilter(Tql())
        self._tql.add_filter('hasIndicator', TqlOperator.EQ, indicators, TqlType.SUB_QUERY)
        return indicators

    @property
    def has_security_label(self):
        """Return **SecurityLabel** for further filtering."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label_filter import SecurityLabelFilter

        security_labels = SecurityLabelFilter(Tql())
        self._tql.add_filter('hasSecurityLabel', TqlOperator.EQ, security_labels, TqlType.SUB_QUERY)
        return security_labels

    @property
    def has_tag(self):
        """Return **TagFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.tags.tag_filter import TagFilter

        tags = TagFilter(Tql())
        self._tql.add_filter('hasTag', TqlOperator.EQ, tags, TqlType.SUB_QUERY)
        return tags

    @property
    def has_victim(self):
        """Return **VictimFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victims.victim_filter import VictimFilter

        victims = VictimFilter(Tql())
        self._tql.add_filter('hasVictim', TqlOperator.EQ, victims, TqlType.SUB_QUERY)
        return victims

    @property
    def has_victim_asset(self):
        """Return **VictimAssetFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset_filter import VictimAssetFilter

        victim_assets = VictimAssetFilter(Tql())
        self._tql.add_filter('hasVictimAsset', TqlOperator.EQ, victim_assets, TqlType.SUB_QUERY)
        return victim_assets

    def host_dns_active(self, operator: Enum, host_dns_active: bool):
        """Filter DNS Active (Host) based on **hostDnsActive** keyword.

        Args:
            operator: The operator enum for the filter.
            host_dns_active: A flag that determines whether or not DNS is active.
        """
        self._tql.add_filter('hostDnsActive', operator, host_dns_active, TqlType.BOOLEAN)

    def host_whois_active(self, operator: Enum, host_whois_active: bool):
        """Filter Whois Active (Host) based on **hostWhoisActive** keyword.

        Args:
            operator: The operator enum for the filter.
            host_whois_active: A flag that determines whether or not whois is active.
        """
        self._tql.add_filter('hostWhoisActive', operator, host_whois_active, TqlType.BOOLEAN)

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the indicator.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def indicator_active(self, operator: Enum, indicator_active: bool):
        """Filter Indicator Status based on **indicatorActive** keyword.

        Args:
            operator: The operator enum for the filter.
            indicator_active: The status (active/inactive) of the indicator.
        """
        self._tql.add_filter('indicatorActive', operator, indicator_active, TqlType.BOOLEAN)

    def last_false_positive(
        self, operator: Enum, last_false_positive: Arrow | datetime | int | str
    ):
        """Filter False Positive Last Observed based on **lastFalsePositive** keyword.

        Args:
            operator: The operator enum for the filter.
            last_false_positive: The date the indicator has been last flagged as a false positive.
        """
        last_false_positive = self.util.any_to_datetime(last_false_positive).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('lastFalsePositive', operator, last_false_positive, TqlType.STRING)

    def last_modified(self, operator: Enum, last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **lastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            last_modified: The date the indicator was last modified in the system.
        """
        last_modified = self.util.any_to_datetime(last_modified).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastModified', operator, last_modified, TqlType.STRING)

    def last_observed(self, operator: Enum, last_observed: Arrow | datetime | int | str):
        """Filter Last Observed based on **lastObserved** keyword.

        Args:
            operator: The operator enum for the filter.
            last_observed: The date the indicator has been last observed.
        """
        last_observed = self.util.any_to_datetime(last_observed).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastObserved', operator, last_observed, TqlType.STRING)

    def observation_count(self, operator: Enum, observation_count: int | list):
        """Filter Observation Count based on **observationCount** keyword.

        Args:
            operator: The operator enum for the filter.
            observation_count: The number of times the indicator has been observed.
        """
        if isinstance(observation_count, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('observationCount', operator, observation_count, TqlType.INTEGER)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the indicator.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name of the indicator.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def rating(self, operator: Enum, rating: int | list):
        """Filter Threat Rating based on **rating** keyword.

        Args:
            operator: The operator enum for the filter.
            rating: The rating of the indicator.
        """
        if isinstance(rating, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('rating', operator, rating, TqlType.INTEGER)

    def security_label(self, operator: Enum, security_label: list | str):
        """Filter Security Label based on **securityLabel** keyword.

        Args:
            operator: The operator enum for the filter.
            security_label: The name of a security label applied to the indicator.
        """
        if isinstance(security_label, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('securityLabel', operator, security_label, TqlType.STRING)

    def source(self, operator: Enum, source: list | str):
        """Filter Source based on **source** keyword.

        Args:
            operator: The operator enum for the filter.
            source: The default source of the indicator.
        """
        if isinstance(source, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('source', operator, source, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The summary of the indicator.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def tag(self, operator: Enum, tag: list | str):
        """Filter Tag based on **tag** keyword.

        Args:
            operator: The operator enum for the filter.
            tag: The name of a tag applied to the indicator.
        """
        if isinstance(tag, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tag', operator, tag, TqlType.STRING)

    def tag_owner(self, operator: Enum, tag_owner: int | list):
        """Filter Tag Owner ID based on **tagOwner** keyword.

        Args:
            operator: The operator enum for the filter.
            tag_owner: The ID of the owner of a tag.
        """
        if isinstance(tag_owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tagOwner', operator, tag_owner, TqlType.INTEGER)

    def tag_owner_name(self, operator: Enum, tag_owner_name: list | str):
        """Filter Tag Owner Name based on **tagOwnerName** keyword.

        Args:
            operator: The operator enum for the filter.
            tag_owner_name: The name of the owner of a tag.
        """
        if isinstance(tag_owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('tagOwnerName', operator, tag_owner_name, TqlType.STRING)

    def threat_assess_score(self, operator: Enum, threat_assess_score: int | list):
        """Filter ThreatAssess Score based on **threatAssessScore** keyword.

        Args:
            operator: The operator enum for the filter.
            threat_assess_score: The threat-assessed score of the indicator.
        """
        if isinstance(threat_assess_score, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('threatAssessScore', operator, threat_assess_score, TqlType.INTEGER)

    def type(self, operator: Enum, type: int | list):  # pylint: disable=redefined-builtin
        """Filter Type ID based on **type** keyword.

        Args:
            operator: The operator enum for the filter.
            type: The ID of the indicator type.
        """
        if isinstance(type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('type', operator, type, TqlType.INTEGER)

    def type_name(self, operator: Enum, type_name: list | str):
        """Filter Type Name based on **typeName** keyword.

        Args:
            operator: The operator enum for the filter.
            type_name: The name of the indicator type.
        """
        if isinstance(type_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('typeName', operator, type_name, TqlType.STRING)

    def value1(self, operator: Enum, value1: list | str):
        """Filter value1 based on **value1** keyword.

        Args:
            operator: The operator enum for the filter.
            value1: No description provided.
        """
        if isinstance(value1, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('value1', operator, value1, TqlType.STRING)

    def value2(self, operator: Enum, value2: list | str):
        """Filter value2 based on **value2** keyword.

        Args:
            operator: The operator enum for the filter.
            value2: No description provided.
        """
        if isinstance(value2, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('value2', operator, value2, TqlType.STRING)

    def value3(self, operator: Enum, value3: list | str):
        """Filter value3 based on **value3** keyword.

        Args:
            operator: The operator enum for the filter.
            value3: No description provided.
        """
        if isinstance(value3, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('value3', operator, value3, TqlType.STRING)

    def vt_last_updated(self, operator: Enum, vt_last_updated: Arrow | datetime | int | str):
        """Filter Virus Total Last Updated based on **vtLastUpdated** keyword.

        Args:
            operator: The operator enum for the filter.
            vt_last_updated: The date the indicator has been looked at with Virus Total.
        """
        vt_last_updated = self.util.any_to_datetime(vt_last_updated).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('vtLastUpdated', operator, vt_last_updated, TqlType.STRING)

    def vt_malicious_count(self, operator: Enum, vt_malicious_count: int | list):
        """Filter Virus Total Malicious Count based on **vtMaliciousCount** keyword.

        Args:
            operator: The operator enum for the filter.
            vt_malicious_count: The number of malicious reports for an indicator from Virus Total.
        """
        if isinstance(vt_malicious_count, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('vtMaliciousCount', operator, vt_malicious_count, TqlType.INTEGER)



"""TcEx Framework Module"""

# standard library
from typing import ClassVar

# third-party
from pydantic import create_model, validator
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.api.tc.util.threat_intel_util import ThreatIntelUtil
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidEntityType
from tcex.input.field_type.tc_entity import TCEntity
from tcex.registry import registry


# pylint: disable=no-self-argument
class IndicatorEntity(TCEntity):
    """Indicator Entity Field (Model) Type"""

    indicator_types: ClassVar[list[str]] = []

    @validator('type', allow_reuse=True)
    def is_empty(cls, value: str, field: ModelField) -> str:
        """Validate that the value is a non-empty string."""
        if isinstance(value, str) and value.replace(' ', '') == '':
            raise InvalidEmptyValue(field_name=field.name)
        return value

    @validator('type', allow_reuse=True)
    def is_type(cls, value: str, field: ModelField) -> str:
        """Validate that the entity is of a specific Indicator type."""
        ti_utils = ThreatIntelUtil(session_tc=registry.session_tc)
        indicator_types = cls.indicator_types or ti_utils.indicator_types
        if value.lower() not in [i.lower() for i in indicator_types]:
            raise InvalidEntityType(
                field_name=field.name, entity_type=str(indicator_types), value=value
            )
        return value


def indicator_entity(
    indicator_types: list[str], model_name: str = 'CustomIndicatorEntity'
) -> type[IndicatorEntity]:
    """Dynamically create a Case Management Entity model."""
    return create_model(
        model_name,
        indicator_types=(ClassVar[list[str]], indicator_types),
        __base__=IndicatorEntity,
    )


AddressEntity = indicator_entity(indicator_types=['Address'])
EmailAddressEntity = indicator_entity(indicator_types=['EmailAddress'])
HostEntity = indicator_entity(indicator_types=['Host'])
FileEntity = indicator_entity(indicator_types=['File'])
UrlEntity = indicator_entity(indicator_types=['URL'])



"""TcEx Framework Module"""
# standard library
import logging

# first-party
from tcex.logger.trace_logger import TraceLogger

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class BaseValueError(ValueError):
    """Raise customized exception."""

    def __init__(self, field_name: str, message: str):
        """Customize the exception message."""
        # when a union of types are provided, this logs for each type that's doesn't match
        _logger.trace(f'Checking value for field {field_name}: {message}')
        super().__init__(message)


class BaseTypeError(TypeError):
    """Raise customized exception."""

    def __init__(self, field_name: str, message: str):
        """Customize the exception message."""
        # when a union of types are provided, this logs for each type that's doesn't match
        _logger.trace(f'Checking type for field {field_name}: {message}')
        super().__init__(message)


class InvalidEmptyValue(BaseValueError):
    """Raise customized exception."""

    def __init__(self, field_name: str):
        """Customize the exception message."""
        super().__init__(field_name, 'an empty value is not allowed for this field')


class InvalidEntityType(BaseTypeError):
    """Raise customized exception."""

    def __init__(self, field_name: str, entity_type: str, value: str):
        """Customize the exception message."""
        super().__init__(field_name, f'{value} is not a {entity_type} type')


class InvalidInput(BaseValueError):
    """Raise customized exception."""

    # pylint: disable=useless-super-delegation
    def __init__(self, field_name: str, error: str):
        """Customize the exception message."""
        super().__init__(field_name, error)


class InvalidIntegerValue(BaseValueError):
    """Raise customized exception."""

    def __init__(self, field_name: str, constraint: int, operation: str):
        """Customize the exception message."""
        super().__init__(field_name, f'value must be {operation} {constraint}')


class InvalidLengthValue(BaseValueError):
    """Raise customized exception."""

    def __init__(self, field_name: str, constraint: int, operation: str):
        """Customize the exception message."""
        super().__init__(field_name, f'value must have a {operation} length of {constraint}')


class InvalidPatternValue(BaseValueError):
    """Raise customized exception."""

    def __init__(self, field_name: str, pattern: str):
        """Customize the exception message."""
        super().__init__(field_name, f'value did not match pattern "{pattern}"')


class InvalidType(BaseTypeError):
    """Raise customized exception."""

    def __init__(self, field_name: str, expected_types: str, provided_type: str):
        """Customize the exception message."""
        super().__init__(field_name, f'{provided_type} is not a {expected_types} type')


class InvalidVariableType(BaseTypeError):
    """Raise customized exception."""

    def __init__(self, field_name: str, expected_type: str, provided_type: str):
        """Customize the exception message."""
        super().__init__(field_name, f'{provided_type} is not a {expected_type} type')



"""TcEx Framework Module"""
# standard library
import datetime
import hashlib
import json
import logging
from abc import ABC
from json import JSONEncoder
from typing import Any, Self

# third-party
from pydantic import BaseModel, PrivateAttr

# first-party
from tcex.logger.trace_logger import TraceLogger

# get tcex logger

_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class CustomJSONEncoder(JSONEncoder):
    """Format object in JSON data."""

    def default(self, o: Any) -> str:
        """Format object"""
        if isinstance(o, BaseModel):
            raise ValueError(
                f'Object of type {type(o)} is not JSON serializable. '
                'Please verify that the object has been converted to a dictionary.'
            )
        if isinstance(o, datetime.date | datetime.datetime):
            return o.isoformat()
        return o


class V3ModelABC(BaseModel, ABC):
    """V3 Base Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _dict_hash: str = PrivateAttr()
    _log = _logger
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)
    id: int | None = None

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(**kwargs)

        # when "id" field is present it indicates that the data was returned from the
        # API, otherwise the assumption is that the developer staged the data during
        # instantiation of the object.
        if kwargs and hasattr(self, 'id') and self.id is None:  # pylint: disable=no-member
            self._staged = True

        # store initial dict hash of model
        self._dict_hash = self.gen_model_hash(self.json(sort_keys=True))

    def _calculate_field_inclusion(
        self, field: str, method: str, mode: str | None, nested: bool, property_: dict, value: Any
    ) -> bool:
        """Return True if the field is calculated to be included."""

        # INDICATOR ID RULE: If an indicator has a ID set, then the indicator fields
        #     cannot be provided (updateable is false) or the API request will fail.
        #     Since Note model CAN update the `text` field we need to specifically state
        #     that this test is for the `Indicator Model` only.
        if (
            self.__config__.title == 'Indicator Model'
            and self.id is not None
            and field
            in [
                'address',
                'file',
                'hostName',
                'ip',
                'md5',
                'sha1',
                'sha256',
                'text',
                'url',
                'value1',
                'value2',
                'value3',
            ]
        ):
            return False

        # MODE DELETE RULE: The "id" should be the only field included when delete mode
        #     is enabled. "relationship" is required for FileActions upon delete mode as well.
        if mode == 'delete' and nested is True and field not in ['id', 'name', 'relationship']:
            return False

        # ID RULE: The "id" should not be included for ANY HTTP method on parent object,
        #     but for nested objects the "id" field should be included when available.
        #     PLAT-4074 - updating nested object using "replace"
        if field == 'id' and nested is True and value:
            return True

        # NESTED RULE: For nested objects the body should use the valid POST fields
        #    instead of the PUT fields if id is not available. This handles including
        #    artifact type, attributes and other fields that are needed on the nested object.
        if method == 'PUT' and nested is True and not self.id:  # pylint: disable=no-member
            method = 'POST'

        # SHARED TYPE RULE: For nested "shared types" objects the body should use the valid
        #    PUT methods instead of the POST fields.  This handles not sending owner field
        #    for tag.
        if method == 'POST' and nested is True and self._shared_type is True:
            method = 'PUT'

        # METHOD RULE: If the current method is in the property "methods" list the
        #     field should be included when available.
        if method in property_.get('methods', []) and (value or value in [0, False]):
            return True

        # DEFAULT RULE -> Fields should not be included unless the match a previous rule.
        return False

    # pylint: disable=too-many-return-statements
    def _calculate_nested_inclusion(self, method: str, mode: str | None, model: Self) -> bool:
        """Return True if the field is calculated to be included.

        Case Management Nested Logic:
        * CM Types (e.g., artifacts, notes, task, etc) -> behavior is APPEND and duplicates will
            be created if resent. Object with an ID should never be sent in parent update. Updating
            the nested object should be done directly on the object.
        * User/User Group Types -> ???
        * Attributes -> behavior uses MODE and supports append, delete, and replace.
        * Tags -> behavior is REPLACE on tags so all tags that need to be on the CM object should
            be include in both POST and PUT.

        Threat Intelligence Nested Logic:
        * Associations (e.g., groups, indicators, victim assets) -> behavior uses MODE and supports
            append, delete, and replace.
        * Attributes -> behavior uses MODE and supports append, delete, and replace.
        * Tags -> behavior uses MODE and supports append, delete, and replace.
        * Security Labels -> behavior uses MODE and supports append, delete, and replace.

        """
        # EMPTY RULE: If there is no value in the model it should NOT be INCLUDED.
        if not model:
            return False

        # METHOD RULE: When HTTP method is DELETE or GET nested objects should NOT be INCLUDED.
        if method in ['DELETE', 'GET']:
            return False

        # STAGED RULE: Any nested object provided by developer should be INCLUDED.
        if model._staged is True:
            return True

        # POST RULE: When method is POST all nested object should be INCLUDED.
        if method == 'POST':
            return True

        # Current Object Restrictions:
        # * The method is PUT
        # * The nested object was NOT added via the stage_xxx method
        # * The nested object contains a ID or Name Field
        # * The nested object could either have been set during initializing the object or fetched.

        # CM and TI endpoint behaves differently. Start with rules based on the parent type,
        # then add more specific rules.

        if self._cm_type is True:
            #
            # CM PARENT TYPES
            #

            # Nested Types:
            # * CM Types (Artifact, Artifact Type, Case, Note, Task, Workflow Event/Template)
            # * Attributes
            # * Group (currently read-only)
            # * Tags
            # * Users

            # Coverage:
            # * Downloaded from API (e.g., case.get(id=123))
            # * Added on instantiation
            # * Added with stage_xxx() method

            if model._cm_type is True:
                # RULE: Short-Circuit Nested CM Types
                # Nested CM types are updated through their direct endpoints and should
                #    never be INCLUDED when updating the parent. For new nested CM types
                #    added with the stage_xxx() method, the STAGED RULE would trigger
                #    before this rule.
                return False

            if model._shared_type is True:
                # RULE: Nested Tags
                # Nested tags on a parent CM type behave as REPLACE mode and need to be
                #     INCLUDED to prevent being removed.
                return True

            # RULE: Nested Attributes w/ APPEND mode
            # Nested attributes on a parent CM type use the mode feature. When the mode
            #     is APPEND and has been UPDATED, then the attributes should be INCLUDED.
            #     For new nested objects added with the stage_xxx() method, the STAGED
            #     RULE would trigger first.
            # A secondary PATTERN consideration is that attributes can be immediately
            #     updated using the attribute.updated() method. While this isn't as
            #     efficient as updating them all in one request, it's is a simpler
            #     development design pattern.

            if mode == 'replace':
                # RULE: Nested Attributes w/ REPLACE mode
                # Nested attributes on a parent CM type use the mode feature. When the mode
                #     is REPLACE the attributes should be INCLUDED.
                return True

            # RULE: Nested Attributes w/ DELETE mode
            # Nested attributes on a parent CM type use the mode feature. When the mode
            #     is DELETE the attribute should NOT be INCLUDED. Any attribute that was
            #     added by the developer using the stage_xxx() method would have hit the
            #     STAGED RULE above and would be INCLUDED.
            # A secondary PATTERN consideration is that attributes can be immediately
            #     deleted using the attribute.delete() method. While this isn't as
            #     efficient as deleting them all in one request, it's is a simpler
            #     development design pattern.

            # All non-matching nested object that did not match a rule above will NOT be INCLUDED.
            return False

        #
        # TI PARENT TYPES (Groups, Indicators, Victim, and Victims Assets)
        #

        # Nested Types:
        # * Associations (Groups, Indicators, Victim Assets)
        # * Attributes
        # * Security Labels
        # * Tags

        # Coverage:
        # * Downloaded from API
        # * Added on instantiation
        # * Added with stage_xxx() method

        if mode == 'append' and self._associated_type:
            # RULE: Nested Object w/ APPEND mode
            # Nested object on a parent CM type use the mode feature. When the mode
            #     is APPEND and not STAGED the object should NOT be INCLUDED.
            return True

        if mode == 'replace':
            # RULE: Nested Object w/ REPLACE mode
            # Nested object on a parent TI type use the mode feature. When the mode
            #     is REPLACE the object should be INCLUDED.
            return True

        # * security_label -> delete (support id or name only)
        # * tag -> delete (support id or name only)
        if (
            mode == 'delete'
            and (model._shared_type is True or self._associated_type is True)
            and (model.id is not None or model.name is not None)  # type: ignore
        ):
            # RULE: Nested Shared Object w/ DELETE mode (TAGS, SECURITY LABELS)
            # Nested shared object on a parent TI type use the mode feature. When the mode
            #     is DELETE the shard object should not be INCLUDED. Any object that was
            #     added by the developer would have hit the STAGED RULE above and would
            #     be INCLUDED.
            return True

        # * associated -> delete (support id only)
        # * attribute -> delete (support id only)
        # RULE: Nested Object w/ DELETE mode
        # Nested object on a parent TI type use the mode feature. When the mode
        #     is DELETE the object should not be INCLUDED. Any object that was
        #     added by the developer would have hit the STAGED RULE above and would
        #     be INCLUDED.

        # All non-matching nested object that did not match a rule above will NOT be INCLUDED.
        return False

    def _process_nested_data_array(self, method: str, mode: str | None, nested_object: Self):
        """Process the nested data object (e.g., GroupsModel.data).

        Nested Object Inclusion Rules:
        * If the method is not PUT.
        * The model doesn't have an "id" field.
          * Object was created using the "add_xxx" method on the parent object.
        * The model's "updated" field is True.
          * Object was modified after it was created. Since object pulled from the API
              are create using **kwargs they are not updated.
        """
        _data = []
        for model in nested_object.data:  # type: ignore
            if self._calculate_nested_inclusion(method, mode, model):
                data = model.gen_body(method, mode, nested=True)
                if data:
                    _data.append(data)

        return _data

    def _process_nested_data_object(self, method: str, mode: str | None, nested_object: Self):
        """Process the nested data object (e.g., GroupsModel.data).

        Nested Object Inclusion Rules:
        * If the method is not PUT.
        * The model doesn't have an "id" field.
          * Object was created using the "add_xxx" method on the parent object.
        * The model's "updated" field is True.
          * Object was modified after it was created. Since object pulled from the API
              are create using **kwargs they are not updated.
        """
        if self._calculate_nested_inclusion(method, mode, nested_object.data):  # type: ignore
            data = nested_object.data.gen_body(method, mode, nested=True)  # type: ignore
            if data:
                return data
        return None

    def _properties(self) -> dict[str, dict[str, str]]:
        """Return properties of the current model."""
        schema = self.schema(by_alias=False)
        if schema.get('properties') is not None:
            return schema.get('properties', {})
        return schema.get('definitions', {}).get(self.__class__.__name__, {}).get('properties', {})

    @staticmethod
    def gen_model_hash(json_: str) -> str:
        """Return the current dict hash."""
        # get hash of dict
        hash_ = hashlib.md5()  # nosec
        encoded = json_.encode()
        hash_.update(encoded)
        return hash_.hexdigest()

    def gen_body(
        self,
        method: str,
        mode: str | None = None,
        exclude_none: bool = True,
        nested: bool = False,
    ) -> dict:
        """Return the generated body.

        The field included in the body depend on the HTTP Method and whether or not the object
        is nested. For example the ID should not be send on the parent object on a POST or PUT,
        but should be added for a PUT on a nested object.
        """
        _body = {}
        schema_properties = self._properties()
        for name, value in self:
            if exclude_none is True and value is None:
                continue

            # get the current field from the schema to us in validating method membership.
            if schema_properties.get(name) is None:
                # a field not being available does not indicate a failure, it could simple
                # be the incorrect field was passed to the object, which will be dropped.
                self._log.warning(
                    f'action=schema-check, type={self.__class__.__name__}, '
                    f'property={name}, result=not-found'
                )
                continue

            property_ = schema_properties.get(name, {})
            key = property_['title']
            if isinstance(value, BaseModel) and property_.get('read_only') is False:
                value: Self = value
                # Handle nested model that should be included in the body (non-read-only).

                if hasattr(value, 'data') and isinstance(value.data, list):  # type: ignore
                    # Handle nested object types where the "data" field contains an array of model.
                    _data = self._process_nested_data_array(method, mode, value)
                    if _data:
                        _body.setdefault(key, {})['data'] = _data

                        # value as a model can be ArtifactTypeModel, ArtifactModel, etc.
                        if value._mode_support:  # type: ignore
                            # Use the default mode defined in the model ("append") or the
                            # mode passed into this method as an override.
                            _body.setdefault(key, {})['mode'] = mode or value.mode  # type: ignore

                elif hasattr(value, 'data'):
                    # Handle "non-standard" condition for Assignee where the nested "data"
                    # field contains an object instead of an Array.
                    _data = self._process_nested_data_object(method, mode, value)
                    if _data:
                        _body.setdefault(key, {})['data'] = _data

                        # Handle the extra "type" field that is only on Assignee objects.
                        if hasattr(value, 'type'):
                            _body.setdefault(key, {})['type'] = value.type  # type: ignore

                else:
                    # Handle nested object types where field contains an object.
                    # (e.g., CaseModel -> workflowTemplate field)
                    # if method == 'POST' or value.id is None or value.updated is True:
                    if self._calculate_nested_inclusion(method, mode, value):
                        _data = value.gen_body(method, mode, nested=True)
                        if _data:
                            _body[key] = _data

            elif self._calculate_field_inclusion(key, method, mode, nested, property_, value):
                # Handle non-nested fields and their values based on well defined rules.
                _body[key] = value

        return _body

    def gen_body_json(
        self,
        method: str,
        mode: str | None = None,
        indent: int | None = None,
        sort_keys: bool = False,
    ) -> str:
        """Wrap gen_body method returning JSON instead of dict."""
        # ensure mode is set to lower case if provided on gen_body entry point
        if mode is not None:
            mode = mode.lower()

        body = self.gen_body(method, mode)
        return json.dumps(
            body,
            cls=CustomJSONEncoder,
            indent=indent,
            sort_keys=sort_keys,
        )

    @property
    def updated(self):
        """Return True if model values have changed, else False."""
        return self._dict_hash != self.gen_model_hash(self.json(sort_keys=True))



"""TcEx Framework Module"""
# standard library
from abc import ABC
from textwrap import TextWrapper

# first-party
from tcex.api.tc.v3._gen._gen_abc import GenerateABC
from tcex.util.string_operation import SnakeString


class GenerateModelABC(GenerateABC, ABC):
    """Generate Models for Case Management Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        self.requirements = {
            'standard library': [],
            'third-party': [
                {'module': 'pydantic', 'imports': ['BaseModel', 'Extra', 'Field']},
            ],
            'first-party': [
                {'module': 'tcex.util', 'imports': ['Util']},
                {'module': 'tcex.api.tc.v3.v3_model_abc', 'imports': ['V3ModelABC']},
            ],
            'first-party-forward-reference': [],
        }
        self.validators = {}

    def _add_module_class(self, from_: str, module: str, class_: str):
        """Add pydantic validator only when required."""
        for lib in self.requirements[from_]:
            if isinstance(lib, dict):
                if lib.get('module') == module:
                    lib['imports'].append(class_)

    def _add_pydantic_validator(self):
        """Add pydantic validator only when required."""
        self._add_module_class('third-party', 'pydantic', 'validator')

    def _add_pydantic_private_attr(self):
        """Add pydantic validator only when required."""
        self._add_module_class('third-party', 'pydantic', 'PrivateAttr')

    def _gen_code_validator_method(self, type_: str, fields: list[str]) -> str:
        """Return the validator code

        @validator('artifact_type', always=True, pre=True)
        def _validate_artifact_type(cls, v):
            if not v:
                return ArtifactTypeModel()
            return v
        """
        type_ = self.util.camel_string(type_)

        fields_string = ', '.join(f'\'{field}\'' for field in fields)
        return '\n'.join(
            [
                f'''{self.i1}@validator({fields_string}, always=True, pre=True)''',
                f'''{self.i1}def _validate_{type_.snake_case()}(cls, v):''',
                f'''{self.i2}if not v:''',
                f'''{self.i3}return {type_}Model()  # type: ignore''',
                f'''{self.i2}return v''',
                '',
            ]
        )

    # TODO: [low] bsummers - research combining this method with parent method
    # pylint: disable=arguments-differ
    def _format_description(self, description: str, length: int) -> str:
        """Format description for field."""
        # fix descriptions coming from core API endpoint
        if description[-1] not in ('.', '?', '!'):
            description += '.'

        # fix core descriptions that are not capitalized.
        description_words = description.split(' ')
        description = f'{description_words[0].title()} ' + ' '.join(description_words[1:])

        # there are 23 characters used on the line (indent -> key -> quote -> comma)
        other_used_space = 22  # space used by indent -> single quote -> key -> comma
        if len(description) < length - other_used_space:
            return f'\'{description}\''

        _description = ['(']
        _description.extend(
            [
                f'{line}\''
                for line in TextWrapper(
                    break_long_words=False,
                    drop_whitespace=False,
                    expand_tabs=True,
                    initial_indent=' ' * 12 + '\'',
                    subsequent_indent=' ' * 12 + '\'',
                    tabsize=15,
                    width=length - 1,
                ).wrap(description)
            ]
        )
        _description.append(f'{self.i2})')
        return '\n'.join(_description)

    def gen_doc_string(self) -> str:
        """Generate doc string."""
        # return (
        #     f'"""{self.type_.singular().title()} / {self.type_.plural().title()} Model"""\n'
        #     '# pylint: disable=no-member,no-self-argument,wrong-import-position\n'
        # )
        return (
            '"""TcEx Framework Module"""\n'
            '# pylint: disable=no-member,no-self-argument,wrong-import-position\n'
        )

    def gen_container_class(self) -> str:
        """Generate the Container Model

        class ArtifactsModel(
            BaseModel,
            title='Artifacts Model',
            alias_generator=Util().snake_to_camel,
            validate_assignment=True,
        ):
        """
        return '\n'.join(
            [
                '',
                f'''class {self.type_.plural().pascal_case()}Model(''',
                f'''{self.i1}BaseModel,''',
                f'''{self.i1}title='{self.type_.plural().pascal_case()} Model',''',
                f'''{self.i1}alias_generator=Util().snake_to_camel,''',
                f'''{self.i1}validate_assignment=True,''',
                '''):''',
                f'''{self.i1}"""{self.type_.plural().title()} Model"""''',
                '',
                '',
            ]
        )

    def gen_container_fields(self) -> str:
        """Generate the Container Model fields

        data: list[ArtifactModel] | None = Field(
            [],
            description='The data of the Cases.',
            methods=['POST', 'PUT'],
            title='data',
        )
        """
        return '\n'.join(
            [
                (
                    f'''{self.i1}data: '''
                    f'''list[{self.type_.singular().pascal_case()}Model] | None '''
                    '''= Field('''
                ),
                f'''{self.i2}[],''',
                (
                    f'''{self.i2}description='The data for the '''
                    f'''{self.type_.plural().pascal_case()}.','''
                ),
                f'''{self.i2}methods=['POST', 'PUT'],''',
                f'''{self.i2}title='data',''',
                f'''{self.i1})''',
                f'''{self.i1}mode: str = Field(''',
                f'''{self.i2}'append',''',
                (
                    f'''{self.i2}description='The PUT mode for nested '''
                    '''objects (append, delete, replace). Default: append','''
                ),
                f'''{self.i2}methods=['POST', 'PUT'],''',
                f'''{self.i2}title='append',''',
                f'''{self.i1})''',
                '',
                '',
            ]
        )

    def gen_container_private_attrs(self) -> str:
        """Generate doc string."""
        self._add_pydantic_private_attr()
        if self.type_.lower() in [
            'case_attributes',
            'group_attributes',
            'groups',
            'indicator_attributes',
            'indicators',
            'security_labels',
            'cases',
            'artifacts',
            'tags',
            # 'task_assignments',
        ]:
            return '\n'.join(
                [
                    f'{self.i1}_mode_support = PrivateAttr(True)',
                    '',
                    '',
                ]
            )
        return '\n'.join(
            [
                f'{self.i1}_mode_support = PrivateAttr(False)',
                '',
                '',
            ]
        )

    def gen_data_class(self) -> str:
        """Generate the Data Class

        class ArtifactDataModel(
            BaseModel,
            title='Artifact Data',
            alias_generator=Util().snake_to_camel,
            validate_assignment=True,
        ):
        """
        return '\n'.join(
            [
                '',
                f'''class {self.type_.singular().pascal_case()}DataModel(''',
                f'''{self.i1}BaseModel,''',
                f'''{self.i1}title='{self.type_.singular().pascal_case()} Data Model',''',
                f'''{self.i1}alias_generator=Util().snake_to_camel,''',
                f'''{self.i1}validate_assignment=True,''',
                '''):''',
                f'''{self.i1}"""{self.type_.plural().title()} Data Model"""''',
                '',
                '',
            ]
        )

    def gen_data_fields(self) -> str:
        """Generate the Data Model fields

        data: 'ArtifactModel | None' = Field(
            None,
            description='The data of the Artifact.',
            title='data',
        )
        """
        return '\n'.join(
            [
                (
                    f'''{self.i1}data: '''
                    f'''list[{self.type_.singular().pascal_case()}Model] | None '''
                    '''= Field('''
                ),
                f'''{self.i2}[],''',
                (
                    f'''{self.i2}description='The data for the '''
                    f'''{self.type_.plural().pascal_case()}.','''
                ),
                f'''{self.i2}methods=['POST', 'PUT'],''',
                f'''{self.i2}title='data',''',
                f'''{self.i1})''',
                '',
                '',
            ]
        )

    def gen_model_class(self) -> str:
        """Generate the model class

        class ArtifactModel(
            BaseModel,
            title='Artifact Model',
            alias_generator=Util().snake_to_camel,
            validate_assignment=True,
        ):
        """
        return '\n'.join(
            [
                '',
                f'''class {self.type_.singular().pascal_case()}Model(''',
                f'''{self.i1}V3ModelABC,''',
                f'''{self.i1}alias_generator=Util().snake_to_camel,''',
                f'''{self.i1}extra=Extra.allow,''',
                f'''{self.i1}title='{self.type_.singular().pascal_case()} Model',''',
                f'''{self.i1}validate_assignment=True,''',
                '''):''',
                f'''{self.i1}"""{self.type_.singular().title()} Model"""''',
                '',
                '',
            ]
        )

    def gen_model_fields(self) -> str:
        """Generate the Model

        Example field_data:
        "analyticsPriority": {
            "readOnly": true,
            "type": "String"
        }

        Example Field:
        analytics_priority: str | None = Field(
            None,
            allow_mutation=False,
            description='The **analytics priority** for the Artifact.',
            read_only=True,
            title='analyticsPriority',
        )
        """
        _model = []
        for prop in self._prop_models:
            # only add requirements for non-current model type
            if prop.extra.type != self.type_.plural().pascal_case():
                if prop.extra.import_data and prop.extra.import_source:
                    self.requirements[prop.extra.import_source].append(prop.extra.import_data)

            # add validator
            if prop.extra.model is not None:
                self.validators.setdefault(prop.type, []).append(prop.name.snake_case())

            # update model
            _model.append(
                f'''{self.i1}{prop.extra.alias.snake_case()}: {prop.extra.typing_type} = Field('''
            )
            _model.append(f'''{self.i2}None,''')  # the default value

            # allow_mutation / readOnly
            if prop.read_only is True and prop.name != 'id':
                _model.append(f'''{self.i2}allow_mutation=False,''')  # readOnly/mutation setting

            # alias
            if prop.extra.alias is not None and prop.name != prop.extra.alias:
                _model.append(f'''{self.i2}alias='{prop.name}',''')

            # applies_to
            if prop.applies_to is not None:
                _model.append(f'''{self.i2}applies_to={prop.applies_to},''')

            # conditional_required
            if prop.conditional_required is not None:
                _model.append(f'''{self.i2}conditional_required={prop.conditional_required},''')

            # description - Use the TC provided description
            # if available, otherwise use a default format.
            field_description = self._format_description(
                (
                    prop.description
                    or (
                        f'The **{prop.name.space_case().lower()}** '
                        f'for the {self.type_.singular().title()}.'
                    )
                ).replace('\'', '\\\''),
                100,
            )
            if field_description is not None:
                _model.append(f'''{self.i2}description={field_description},''')

            # methods (HTTP)
            if prop.extra.methods:
                _model.append(f'''{self.i2}methods={prop.extra.methods},''')

            # max_length
            # if prop.max_length is not None:
            #     _model.append(f'''{self.i2}max_length={prop.max_length},''')

            # max_size
            # if field_max_size is not None:
            #     _model.append(f'''{self.i2}max_items={field_max_size},''')

            # max_value
            if prop.max_value is not None:
                _model.append(f'''{self.i2}maximum={prop.max_value},''')

            # min_length
            # if prop.min_length is not None:
            #     _model.append(f'''{self.i2}min_length={prop.min_length},''')

            # min_value
            if prop.min_value is not None:
                _model.append(f'''{self.i2}minimum={prop.min_value},''')

            # readOnly/allow_mutation setting
            _model.append(f'''{self.i2}read_only={prop.read_only},''')

            # required-alt-field
            if prop.required_alt_field is not None:
                _model.append(f'''{self.i2}required_alt_field='{prop.required_alt_field}',''')

            # title
            _model.append(f'''{self.i2}title='{prop.name}',''')

            _model.append(f'''{self.i1})''')
        _model.append('')

        return '\n'.join(_model)

    def gen_model_private_attrs(self) -> str:
        """Generate doc string."""
        associated_type = False
        cm_type = False
        shared_type = False

        self._add_pydantic_private_attr()
        if self.type_.lower() in [
            'artifacts',
            'artifact_types',
            'cases',
            'notes',
            'tasks',
            'workflow_events',
            'workflow_templates',
        ]:
            cm_type = True
        elif self.type_.lower() in [
            'security_labels',
            'tags',
        ]:
            shared_type = True
        elif self.type_.lower() in [
            'groups',
            'indicators',
            'victim_assets',
        ]:
            associated_type = True

        return '\n'.join(
            [
                f'{self.i1}_associated_type = PrivateAttr({associated_type})',
                f'{self.i1}_cm_type = PrivateAttr({cm_type})',
                f'{self.i1}_shared_type = PrivateAttr({shared_type})',
                f'{self.i1}_staged = PrivateAttr(False)',
                '',
                '',
            ]
        )

    def gen_requirements_first_party_forward_reference(self):
        """Generate first-party forward reference, imported at the bottom of the file."""
        _libs = []
        for from_, libs in self.requirements.items():
            if from_ not in ['first-party-forward-reference']:
                continue

            if libs:
                _libs.append('')  # add newline
                _libs.append('')  # add newline
                _libs.append('# first-party')
                for lib in sorted(libs):
                    _libs.append(lib)
        return '\n'.join(_libs)

    def gen_forward_reference(self):
        """Generate first-party forward reference, imported at the bottom of the file."""
        return '\n'.join(
            [
                '',
                '',
                '# add forward references',
                f'{self.type_.singular().pascal_case()}DataModel.update_forward_refs()',
                f'{self.type_.singular().pascal_case()}Model.update_forward_refs()',
                f'{self.type_.plural().pascal_case()}Model.update_forward_refs()',
                '',
            ]
        )

    def gen_validator_methods(self) -> str:
        """Generate model validator."""
        _v = []

        validators = dict(
            sorted({k: v for k, v in self.validators.items() if v is not None}.items())
        )
        for key, fields in validators.items():
            if not fields:
                continue
            _v.append(self._gen_code_validator_method(key, fields))

        # add blank line above validators only if validators exists
        if _v:
            _v.insert(0, '')
            self._add_pydantic_validator()

        return '\n'.join(_v)



"""TcEx Framework Module"""

# third-party
from pydantic import Extra

# first-party
from tcex.input.model.api_model import ApiModel
from tcex.input.model.proxy_model import ProxyModel


class ModuleRequestsSessionModel(ApiModel, ProxyModel):
    """Model Definition

    This model provides all the inputs required by the "tcex.requests_session" module.
    """

    class Config:
        """Model Config"""

        extra = Extra.ignore



"""TcEx Framework Module"""
# standard library
from enum import Enum, unique


@unique
class TqlType(Enum):
    """ThreatConnect API TQL Types"""

    STRING = 'String'
    INTEGER = 'Integer'
    BOOLEAN = 'Boolean'
    SUB_QUERY = 'Sub Query'



"""TcEx Framework Module"""
# pylint: disable=no-self-argument
# standard library
from typing import Any

# third-party
from pydantic import BaseModel, Extra, root_validator, validator

# first-party
from tcex.app.config import InstallJson

# get instance of InstallJson
ij = InstallJson()


class CreateConfigModel(BaseModel):
    """Create Config Model"""

    tc_playbook_out_variables: list[str]
    trigger_id: int

    @validator('tc_playbook_out_variables', pre=True)
    def _tc_playbook_out_variables(cls, v) -> list[str]:
        """Convert tc_playbook_out_variables into a list of strings.

        This value comes-in as a comma-separated list.
        """
        return v.split(',') if v else []

    # TODO: [low] workaround for PLAT-4393
    @root_validator(pre=True)
    def _update_inputs(cls, values: dict[str, Any]):
        """Convert empty strings to None.

        Workarounds for core issues:
            - Core sends '' for field that are not populated instead of sending a null value.
            - Core sends a string for multi-value inputs instead of an array.
        """
        for field, value in values.items():
            param = ij.model.get_param(field)
            if param is not None and (param.type.lower() == 'multichoice' or param.allow_multiple):
                if value is not None and not isinstance(value, list):
                    values[field] = value.split(ij.model.list_delimiter or '|')

            if value == '':
                values[field] = None

        return values

    class Config:
        """Model Config"""

        extra = Extra.allow



"""TcEx Framework Module"""
# standard library
import gzip
import json
import os
import shelve  # nosec
import sys
import threading
import time
import traceback
from collections import deque
from collections.abc import Callable
from typing import Any

# third-party
from requests import Response, Session

# first-party
from tcex.api.tc.v2.batch.batch_submit import BatchSubmit
from tcex.api.tc.v2.batch.batch_writer import BatchWriter, GroupType, IndicatorType
from tcex.exit.error_code import handle_error
from tcex.input.input import Input


class Batch(BatchWriter, BatchSubmit):
    """ThreatConnect Batch Import Module

    Args:
        inputs: The App inputs.
        session_tc: The ThreatConnect API session.
        owner: The ThreatConnect owner for Batch action.
        action: Action for the batch job ['Create', 'Delete'].
        attribute_write_type: Write type for attributes ['Append', 'Replace'].
        halt_on_error: If True, any batch error will halt the batch job.
        playbook_triggers_enabled: If True, Playbook will be triggered when TI data is created.
        security_label_write_type: Write type for labels ['Append', 'Replace'].
        tag_write_type: Write type for tags ['Append', 'Replace'].
    """

    def __init__(
        self,
        inputs: Input,
        session_tc: Session,
        owner: str,
        action: str = 'Create',
        attribute_write_type: str = 'Replace',
        halt_on_error: bool = True,
        playbook_triggers_enabled: bool = False,
        tag_write_type: str = 'Replace',
        security_label_write_type: str = 'Replace',
    ):
        """Initialize instance properties."""
        BatchWriter.__init__(self, inputs=inputs, session_tc=session_tc, output_dir='')
        BatchSubmit.__init__(
            self,
            inputs=inputs,
            session_tc=session_tc,
            owner=owner,
            action=action,
            attribute_write_type=attribute_write_type,
            halt_on_error=halt_on_error,
            playbook_triggers_enabled=playbook_triggers_enabled,
            tag_write_type=tag_write_type,
            security_label_write_type=security_label_write_type,
        )

        self._action = action
        self._attribute_write_type = attribute_write_type
        self._halt_on_error = halt_on_error
        self._owner = owner
        self._playbook_triggers_enabled = playbook_triggers_enabled
        self._security_label_write_type = security_label_write_type
        self._tag_write_type = tag_write_type

        # properties
        self._batch_max_chunk = 5_000
        self._batch_max_size = 75_000_000  # max size in bytes
        self._file_merge_mode = None
        self._file_threads = []
        self._hash_collision_mode = None
        self._submit_thread = None

        # global overrides on batch/file errors
        self._halt_on_batch_error = None
        self._halt_on_file_error = None
        self._halt_on_poll_error = None

        # debug/saved flags
        self._saved_xids = None
        self._saved_groups = None  # indicates groups shelf file was provided
        self._saved_indicators = None  # indicates indicators shelf file was provided
        self.enable_saved_file = False

        # default properties
        self._batch_data_count = None
        self._poll_interval = None
        self._poll_interval_times = []
        self._poll_timeout = 3600

        # batch debug/replay variables
        self._debug = None
        self.debug_path = os.path.join(self.inputs.model.tc_temp_path, 'DEBUG')
        self.debug_path_batch = os.path.join(self.debug_path, 'batch_data')
        self.debug_path_group_shelf = os.path.join(self.debug_path, 'groups-saved')
        self.debug_path_indicator_shelf = os.path.join(self.debug_path, 'indicators-saved')
        self.debug_path_files = os.path.join(self.debug_path, 'batch_files')
        self.debug_path_xids = os.path.join(self.debug_path, 'xids-saved')

    def _group(self, group_data: dict | GroupType, store: bool = True) -> dict | GroupType:
        """Return previously stored group or new group.

        Args:
            group_data: An Group dict or instance of GroupType.
            store: If True the group data will be stored in instance list.

        Returns:
            (dict|GroupType): The new Group dict/GroupType or the previously stored dict/GroupType.
        """
        if store is False:
            return group_data

        if isinstance(group_data, dict):
            # get xid from dict
            xid = group_data['xid']
        else:
            # get xid from GroupType
            xid = group_data.xid

        if self.groups.get(xid) is not None:
            # return existing group from memory
            group_data = self.groups[xid]
        elif self.groups_shelf.get(xid) is not None:
            # return existing group from shelf
            group_data = self.groups_shelf[xid]
        else:
            # store new group
            self.groups[xid] = group_data
        return group_data

    def _indicator(
        self, indicator_data: dict | IndicatorType, store: bool = True
    ) -> dict | IndicatorType:
        """Return previously stored indicator or new indicator.

        Args:
            indicator_data: An Indicator dict or instance of IndicatorType.
            store: If True the indicator data will be stored in instance list.
        """
        if store is False:
            return indicator_data

        if isinstance(indicator_data, dict):
            # get xid from dict
            xid = indicator_data['xid']
        else:
            # get xid from IndicatorType
            xid = indicator_data.xid

        if self.indicators.get(xid) is not None:
            # return existing indicator from memory
            indicator_data = self.indicators[xid]
        elif self.indicators_shelf.get(xid) is not None:
            # return existing indicator from shelf
            indicator_data = self.indicators_shelf[xid]
        else:
            # store new indicators
            self.indicators[xid] = indicator_data
        return indicator_data

    def close(self):
        """Cleanup batch job."""
        # allow pol thread to complete before wrapping up
        if self._submit_thread and hasattr(self._submit_thread, 'is_alive'):
            self._submit_thread.join()

        # allow file threads to complete before wrapping up job
        for t in self._file_threads:
            t.join()

        self.groups_shelf.close()
        self.indicators_shelf.close()
        if not self.debug and not self.enable_saved_file:
            # delete saved files
            if os.path.isfile(self.group_shelf_fqfn):
                os.remove(self.group_shelf_fqfn)
            if os.path.isfile(self.group_shelf_fqfn):
                os.remove(self.indicator_shelf_fqfn)

    @property
    def data(self) -> dict:
        """Return the batch indicator/group and file data to be sent to the ThreatConnect API.

        **Processing Order:**
        * Process groups in memory up to max batch size.
        * Process groups in shelf to max batch size.
        * Process indicators in memory up to max batch size.
        * Process indicators in shelf up to max batch size.

        This method will remove the group/indicator from memory and/or shelf.

        Returns:
            dict: A dictionary of group, indicators, and/or file data.
        """
        data = {'file': {}, 'group': [], 'indicator': []}
        tracker = {'count': 0, 'bytes': 0}

        # process group from memory, returning if max values have been reached
        if self.data_groups(data, self.groups, tracker) is True:
            return data

        # process group from shelf file, returning if max values have been reached
        if self.data_groups(data, self.groups_shelf, tracker) is True:
            return data

        # process indicator from memory, returning if max values have been reached
        if self.data_indicators(data, self.indicators, tracker) is True:
            return data

        # process indicator from shelf file, returning if max values have been reached
        if self.data_indicators(data, self.indicators_shelf, tracker) is True:
            return data

        return data

    def data_group_association(self, data: dict, tracker: dict, xid: str):
        """Return group dict array following all associations.

        The *data* dict is passed by reference to make it easier to update both the group data
        and file data inline versus passing the data all the way back up to the calling methods.

        Args:
            data: The data dict to update with group and file data.
            tracker: A dict containing total count of all entities collected and
                the total size in bytes of all entities collected.
            xid: The xid of the group to retrieve associations.
        """
        xids = deque()
        xids.append(xid)

        while xids:
            xid = xids.popleft()  # remove current xid
            group_data = None

            if xid in self.groups:
                group_data = self.groups.get(xid)
                del self.groups[xid]
            elif xid in self.groups_shelf:
                group_data = self.groups_shelf.get(xid)
                del self.groups_shelf[xid]

            if group_data:
                file_data, group_data = self.data_group_type(group_data)
                data['group'].append(group_data)
                if file_data:
                    data['file'][xid] = file_data

                # update entity trackers
                tracker['count'] += 1
                tracker['bytes'] += sys.getsizeof(json.dumps(group_data))

                # extend xids with any groups associated with the same GroupType
                xids.extend(group_data.get('associatedGroupXid', []))

    @staticmethod
    def data_group_type(group_data: dict | GroupType) -> tuple[dict, dict]:
        """Return dict representation of group data and file data.

        Args:
            group_data: The group data dict or GroupType.

        Returns:
            Tuple[dict, dict]: A tuple containing file_data and group_data.
        """
        file_data = {}
        if isinstance(group_data, dict):
            # process file content
            file_content = group_data.pop('fileContent', None)
            if file_content is not None:
                file_data = {
                    'fileContent': file_content,
                    'fileName': group_data.get('fileName'),
                    'type': group_data.get('type'),
                }
        else:
            # get the file data from the GroupType and return dict format of GroupType
            if group_data.data.get('type') in ['Document', 'Report']:
                file_data = group_data.file_data
            group_data = group_data.data

        return file_data, group_data

    def data_groups(self, data: dict, groups: dict | shelve.Shelf[Any], tracker: dict) -> bool:
        """Process Group data.

        Args:
            data: The data dict to update with group and file data.
            groups: The list of groups to process.
            tracker: A dict containing total count of all entities collected and
                the total size in bytes of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # convert groups.keys() to a list to prevent dictionary change error caused by
        # the data_group_association function deleting items from the GroupType.

        # process the group
        for xid in list(groups.keys()):
            # get association from group data
            self.data_group_association(data, tracker, xid)

            if tracker['count'] % 2_500 == 0:
                # log count/size at a sane level
                self.log.info(
                    '''feature=batch, action=data-groups, '''
                    f'''count={tracker.get('count'):,}, bytes={tracker.get('bytes'):,}'''
                )

            if (
                tracker['count'] >= self._batch_max_chunk
                or tracker['bytes'] >= self._batch_max_size
            ):
                # stop processing xid once max limit are reached
                self.log.info(
                    '''feature=batch, event=max-value-reached, '''
                    f'''count={tracker.get('count'):,}, bytes={tracker.get('bytes'):,}'''
                )
                return True
        return False

    def data_indicators(
        self, data: dict, indicators: dict | shelve.Shelf[Any], tracker: dict
    ) -> bool:
        """Process Indicator data.

        Args:
            data: The data dict to update with group and file data.
            indicators: The list of indicators to process.
            tracker: A dict containing total count of all entities collected and
                the total size in bytes of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # process the indicators
        for xid, indicator_data in list(indicators.items()):
            if not isinstance(indicator_data, dict):
                indicator_data = indicator_data.data
            data['indicator'].append(indicator_data)
            del indicators[xid]

            # update entity trackers
            tracker['count'] += 1
            tracker['bytes'] += sys.getsizeof(json.dumps(indicator_data))

            if tracker['count'] % 2_500 == 0:
                # log count/size at a sane level
                self.log.info(
                    '''feature=batch, action=data-indicators, '''
                    f'''count={tracker.get('count'):,}, bytes={tracker.get('bytes'):,}'''
                )

            if (
                tracker['count'] >= self._batch_max_chunk
                or tracker['bytes'] >= self._batch_max_size
            ):
                # stop processing xid once max limit are reached
                self.log.info(
                    '''feature=batch, event=max-value-reached, '''
                    f'''count={tracker.get('count'):,}, bytes={tracker.get('bytes'):,}'''
                )
                return True
        return False

    @property
    def debug(self):
        """Return debug setting"""
        if self._debug is None:
            self._debug = False

            # switching DEBUG file to a directory
            if os.path.isfile(self.debug_path):
                os.remove(self.debug_path)
                os.makedirs(self.debug_path, exist_ok=True)

            if os.path.isdir(self.debug_path) and os.access(self.debug_path, os.R_OK):
                # create directories only required when debug is enabled
                # batch_json - store the batch*.json files
                # documents - store the file downloads (e.g., *.pdf)
                # reports - store the file downloads (e.g., *.pdf)
                os.makedirs(self.debug_path, exist_ok=True)
                os.makedirs(self.debug_path_batch, exist_ok=True)
                os.makedirs(self.debug_path_files, exist_ok=True)
                self._debug = True
        return self._debug

    @property
    def halt_on_file_error(self) -> bool:
        """Return halt on file post error value."""
        return self._halt_on_file_error or False

    @halt_on_file_error.setter
    def halt_on_file_error(self, value: bool):
        """Set halt on file post error value."""
        if isinstance(value, bool):
            self._halt_on_file_error = value

    def process_all(self, process_files: bool = True):
        """Process Batch request to ThreatConnect API.

        Args:
            process_files: Send any document or report attachments to the API.
        """
        while True:
            content = self.data
            file_data = content.pop('file', {})
            if not content.get('group') and not content.get('indicator'):
                break

            # special code for debugging App using batchV2.
            self.write_batch_json(content)

            # store the length of the batch data to use for poll interval calculations
            self.log.info(
                '''feature=batch, event=process-all, type=group, '''
                f'''count={len(content['group']):,}'''
            )
            self.log.info(
                '''feature=batch, event=process-all, type=indicator, '''
                f'''count={len(content['indicator']):,}'''
            )

        if process_files:
            self.process_files(file_data)

    def process_files(self, file_data: dict):
        """Process Files for Documents and Reports to ThreatConnect API.

        Args:
            file_data: The file data to be processed.
        """
        for xid, content_data in list(file_data.items()):
            del file_data[xid]  # win or loose remove the entry

            # define the saved filename
            api_branch = 'documents' if content_data.get('type') == 'Report' else 'reports'
            fqfn = os.path.join(
                self.debug_path_files,
                f'''{api_branch}--{xid}--{content_data.get('fileName').replace('/', ':')}''',
            )

            # used for debug/testing to prevent upload of previously uploaded file
            if self.debug and xid in self.saved_xids:
                self.log.debug(
                    f'feature=batch-submit-files, action=skip-previously-saved-file, xid={xid}'
                )
                continue

            if os.path.isfile(fqfn):
                self.log.debug(
                    f'feature=batch-submit-files, action=skip-previously-saved-file, xid={xid}'
                )
                continue

            # process the file content
            content = content_data.get('fileContent')
            if callable(content):
                content_callable_name = getattr(content, '__name__', repr(content))
                self.log.trace(
                    f'feature=batch-submit-files, method={content_callable_name}, xid={xid}'
                )
                content = content_data.get('fileContent')(xid)

            if content is None:
                self.log.warning(f'feature=batch-submit-files, xid={xid}, event=content-null')
                continue

            # write the file to disk
            with open(fqfn, 'wb') as fh:
                fh.write(content)

    @property
    def saved_groups(self) -> bool:
        """Return True if saved group files exits, else False."""
        if self._saved_groups is None:
            self._saved_groups = False
            if (
                self.enable_saved_file
                and os.path.isfile(self.debug_path_group_shelf)
                and os.access(self.debug_path_group_shelf, os.R_OK)
            ):
                self._saved_groups = True
                self.log.debug('feature=batch, event=saved-groups-file-found')
        return self._saved_groups

    @property
    def saved_indicators(self) -> bool:
        """Return True if saved indicators files exits, else False."""
        if self._saved_indicators is None:
            self._saved_indicators = False
            if (
                self.enable_saved_file
                and os.path.isfile(self.debug_path_indicator_shelf)
                and os.access(self.debug_path_indicator_shelf, os.R_OK)
            ):
                self._saved_indicators = True
                self.log.debug('feature=batch, event=saved-indicator-file-found')
        return self._saved_indicators

    @property
    def saved_xids(self) -> list:
        """Return previously saved xids."""
        if self._saved_xids is None:
            self._saved_xids = []
            if self.debug:
                if os.path.isfile(self.debug_path_xids) and os.access(
                    self.debug_path_xids, os.R_OK
                ):
                    with open(self.debug_path_xids) as fh:
                        self._saved_xids = fh.read().splitlines()
        return self._saved_xids

    @saved_xids.setter
    def saved_xids(self, xid: str):
        """Append xid to xids saved file."""
        with open(self.debug_path_xids, 'a') as fh:
            fh.write(f'{xid}\n')

    def submit(  # pylint: disable=arguments-differ, arguments-renamed
        self,
        poll: bool = True,
        errors: bool = True,
        process_files: bool = True,
        halt_on_error: bool = True,
    ) -> dict:
        """Submit Batch request to ThreatConnect API.

        By default this method will submit the job request and data and if the size of the data
        is below the value **synchronousBatchSaveLimit** set in System Setting it will process
        the request synchronously and return the batch status.  If the size of the batch is greater
        than the value set the batch job will be queued.
        Errors are not retrieve automatically and need to be enabled.

        If any of the submit, poll, or error methods fail the entire submit will halt at the point
        of failure. The behavior can be changed by setting halt_on_error to False.

        Each of these methods can also be called on their own for greater control of the submit
        process.

        Args:
            poll: If True poll batch for status.
            errors: If True retrieve any batch errors (only if poll is True).
            process_files: If true send any document or report attachments to the API.
            halt_on_error: If True any exception will raise an error.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        # get file, group, and indicator data
        content = self.data

        # pop any file content to pass to submit_files
        file_data = content.pop('file', {})
        batch_data = (
            self.submit_create_and_upload(content=content, halt_on_error=halt_on_error)
            .get('data', {})
            .get('batchStatus', {})
        )
        batch_id = batch_data.get('id')
        if batch_id is not None:
            self.log.info(f'feature=batch, event=submit, batch-id={batch_id}')
            # job hit queue
            if poll:
                # poll for status
                batch_data = (
                    self.poll(batch_id=batch_id, halt_on_error=halt_on_error)
                    .get('data', {})
                    .get('batchStatus', {})
                )
                if errors:
                    # retrieve errors
                    error_groups = batch_data.get('errorGroupCount', 0)
                    error_indicators = batch_data.get('errorIndicatorCount', 0)
                    if error_groups > 0 or error_indicators > 0:
                        batch_data['errors'] = self.errors(batch_id)
            else:
                # can't process files if status is unknown (polling must be enabled)
                process_files = False

        if process_files:
            # submit file data after batch job is complete
            self._file_threads.append(
                self.submit_thread(
                    name='submit-files',
                    target=self.submit_files,
                    args=(
                        file_data,
                        halt_on_error,
                    ),
                )
            )
        return batch_data

    def submit_all(
        self,
        poll: bool = True,
        errors: bool = True,
        process_files: bool = True,
        halt_on_error: bool = True,
    ) -> list[dict]:
        """Submit Batch request to ThreatConnect API.

        By default this method will submit the job request and data and if the size of the data
        is below the value **synchronousBatchSaveLimit** set in System Setting it will process
        the request synchronously and return the batch status.  If the size of the batch is greater
        than the value set the batch job will be queued.
        Errors are not retrieve automatically and need to be enabled.

        If any of the submit, poll, or error methods fail the entire submit will halt at the point
        of failure. The behavior can be changed by setting halt_on_error to False.

        Each of these methods can also be called on their own for greater control of the submit
        process.

        Args:
            poll: If True poll batch for status.
            errors: If True retrieve any batch errors (only if poll is True).
            process_files: If true send any document or report attachments to the API.
            halt_on_error: If True any exception will raise an error.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        batch_data_array = []
        file_data = {}
        while True:
            batch_data: dict[str, int | list | str] | None = {}
            batch_id: int | None = None

            # get file, group, and indicator data
            content = self.data

            # break loop when end of data is reached
            if not content.get('group') and not content.get('indicator'):
                break

            if self.action.lower() == 'delete':
                # no need to process files on a delete batch job
                process_files = False

                # while waiting of FR for delete support in createAndUpload submit delete request
                # the old way (submit job + submit data), still using V2.
                if len(content) > 0:  # pylint: disable=len-as-condition
                    batch_id = self.submit_job(halt_on_error)
                    if batch_id is not None:
                        batch_data = self.submit_data(
                            batch_id=batch_id, content=content, halt_on_error=halt_on_error
                        )
                else:
                    batch_data = {}
            else:
                # pop any file content to pass to submit_files
                file_data = content.pop('file', {})
                batch_data = (
                    self.submit_create_and_upload(content=content, halt_on_error=halt_on_error)
                    .get('data', {})
                    .get('batchStatus', {})
                )
                batch_id = batch_data.get('id')  # type: ignore

            if batch_id is not None:
                self.log.info(f'feature=batch, event=status, batch-id={batch_id}')
                # job hit queue
                if poll:
                    # poll for status
                    batch_data = (
                        self.poll(batch_id, halt_on_error=halt_on_error)
                        .get('data', {})
                        .get('batchStatus', {})
                    )
                    if errors and batch_data is not None:
                        # retrieve errors
                        error_count = batch_data.get('errorCount', 0)
                        error_groups = batch_data.get('errorGroupCount', 0)
                        error_indicators = batch_data.get('errorIndicatorCount', 0)
                        if (
                            isinstance(error_count, int)
                            and isinstance(error_groups, int)
                            and isinstance(error_indicators, int)
                        ):
                            if error_count > 0 or error_groups > 0 or error_indicators > 0:
                                batch_data['errors'] = self.errors(batch_id)
                else:
                    # can't process files if status is unknown (polling must be enabled)
                    process_files = False

            if process_files:
                # submit file data after batch job is complete
                self._file_threads.append(
                    self.submit_thread(
                        name='submit-files',
                        target=self.submit_files,
                        args=(
                            file_data,
                            halt_on_error,
                        ),
                    )
                )
            batch_data_array.append(batch_data)

            # write errors for debugging
            if isinstance(batch_data, dict):
                batch_errors = batch_data.get('errors', [])
                if isinstance(batch_errors, list) and len(batch_errors) > 0:
                    self.write_error_json(batch_errors)

        return batch_data_array

    def submit_callback(
        self,
        callback: Callable[..., Any],
        content: dict | None = None,
        halt_on_error: bool = True,
    ) -> bool:
        """Submit batch data to ThreatConnect and poll in a separate thread.

        The "normal" submit methods run in serial which will block when the batch poll is running.
        Using this method the submit is done in serial, but the poll method is run in a thread,
        which should allow the App to continue downloading and processing data while the batch
        poll process is running. Only one batch submission is allowed at a time so that any
        critical errors returned from batch can be handled before submitting a new batch job.

        Args:
            callback: The callback method that will handle
                the batch status when polling is complete.
            content: The dict of groups and indicator data (e.g., {"group": [], "indicator": []}).
            halt_on_error: If True the process should halt if any errors are encountered.

        Raises:
            RuntimeError: Raised on invalid callback method.

        Returns:
            bool: False when there is not data to process, else True
        """
        # user provided content or grab content from local group/indicator lists
        if content is not None:
            # process content
            pass
        else:
            content = self.data
        file_data = content.pop('file', {})

        # return False when end of data is reached
        if not content.get('group') and not content.get('indicator'):
            return False

        # block here is there is already a batch submission being processed
        if self._submit_thread and hasattr(self._submit_thread, 'is_alive'):
            self.log.info(
                'feature=batch, event=progress, status=blocked, '
                f'is-alive={self._submit_thread.is_alive()}'
            )
            self._submit_thread.join()
            self.log.debug(
                'feature=batch, event=progress, status=released, '
                f'is-alive={self._submit_thread.is_alive()}'
            )

        # submit the data and collect the response
        batch_data: dict = (
            self.submit_create_and_upload(content=content, halt_on_error=halt_on_error)
            .get('data', {})
            .get('batchStatus', {})
        )
        self.log.trace(f'feature=batch, event=submit-callback, batch-data={batch_data}')

        # launch batch polling in a thread
        self._submit_thread = self.submit_thread(
            name='submit-poll',
            target=self.submit_callback_thread,
            args=(batch_data, callback, file_data),
        )

        return True

    def submit_callback_thread(
        self,
        batch_data: dict,
        callback: Callable[..., Any],
        file_data: dict,
        halt_on_error: bool = True,
    ):
        """Submit data in a thread."""
        batch_id = batch_data.get('id')
        self.log.info(f'feature=batch, event=progress, batch-id={batch_id}')
        if batch_id:
            # when batch_id is None it indicates that batch submission was small enough to be
            # processed inline (without being queued)

            # poll for status
            batch_status = (
                self.poll(batch_id, halt_on_error=halt_on_error).get('data', {}).get('batchStatus')
            )

            # retrieve errors
            if batch_status is not None:
                error_count = batch_status.get('errorCount', 0)
                error_groups = batch_status.get('errorGroupCount', 0)
                error_indicators = batch_status.get('errorIndicatorCount', 0)
                if error_count > 0 or error_groups > 0 or error_indicators > 0:
                    batch_status['errors'] = self.errors(batch_id)
        else:
            batch_status = batch_data

        # launch file upload in a thread *after* batch status is returned. while only one batch
        # submission thread is allowed, there is no limit on file upload threads. the upload
        # status returned by file upload will be ignored when running in a thread.
        if file_data:
            self._file_threads.append(
                self.submit_thread(
                    name='submit-files',
                    target=self.submit_files,
                    args=(
                        file_data,
                        halt_on_error,
                    ),
                )
            )

        # send batch_status to callback
        if callable(callback):
            self.log.debug('feature=batch, event=calling-callback')
            try:
                callback(batch_status)
            except Exception as e:
                self.log.warning(f'feature=batch, event=callback-error, err="""{e}"""')

    def submit_create_and_upload(self, content: dict, halt_on_error: bool = True) -> dict:
        """Submit Batch request to ThreatConnect API.

        Args:
            content: The dict of groups and indicator data.
            halt_on_error: If True the process should halt if any errors are encountered.

        Returns.
            dict: The Batch Status from the ThreatConnect API.
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        # special code for debugging App using batchV2.
        self.write_batch_json(content)

        # store the length of the batch data to use for poll interval calculations
        self.log.info(
            '''feature=batch, event=submit-create-and-upload, type=group, '''
            f'''count={len(content['group']):,}'''
        )
        self.log.info(
            '''feature=batch, event=submit-create-and-upload, type=indicator, '''
            f'''count={len(content['indicator']):,}'''
        )

        try:
            files = (('config', json.dumps(self.settings)), ('content', json.dumps(content)))
            params = {'includeAdditional': 'true'}
            r = self.session_tc.post('/v2/batch/createAndUpload', files=files, params=params)
            if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
                handle_error(
                    code=10510,
                    message_values=[r.status_code, r.text],
                    raise_error=halt_on_error,
                )
            return r.json()
        except Exception as e:
            handle_error(code=10505, message_values=[e], raise_error=halt_on_error)

        return {}

    def submit_files(self, file_data: dict, halt_on_error: bool = True) -> list[dict] | None:
        """Submit Files for Documents and Reports to ThreatConnect API.

        Critical Errors

        * There is insufficient document storage allocated to this account.

        Args:
            halt_on_error: If True any exception will raise an error.
            file_data: The file data to be submitted.

        Returns:
            dict: The upload status for each xid.
        """
        # check global setting for override
        if self.halt_on_file_error is not None:
            halt_on_error = self.halt_on_file_error

        upload_status = []
        self.log.info(f'feature=batch, action=submit-files, count={len(file_data)}')
        for xid, content_data in list(file_data.items()):
            del file_data[xid]  # win or loose remove the entry
            status = True

            # used for debug/testing to prevent upload of previously uploaded file
            if self.debug and xid in self.saved_xids:
                self.log.debug(
                    f'feature=batch-submit-files, action=skip-previously-saved-file, xid={xid}'
                )
                continue

            # process the file content
            content = content_data.get('fileContent')
            if callable(content):
                try:
                    content_callable_name = getattr(content, '__name__', repr(content))
                    self.log.trace(
                        f'feature=batch-submit-files, method={content_callable_name}, xid={xid}'
                    )
                    content = content_data.get('fileContent')(xid)
                except Exception as e:
                    self.log.warning(f'feature=batch, event=file-download-exception, err="""{e}"""')

            if content is None:
                upload_status.append({'uploaded': False, 'xid': xid})
                self.log.warning(f'feature=batch-submit-files, xid={xid}, event=content-null')
                continue

            api_branch = 'documents'
            if content_data.get('type') == 'Report':
                api_branch = 'reports'

            if self.debug and content_data.get('fileName'):
                # special code for debugging App using batchV2.
                fqfn = os.path.join(
                    self.debug_path_files,
                    f'''{api_branch}--{xid}--{content_data.get('fileName').replace('/', ':')}''',
                )
                if os.path.isdir(os.path.dirname(fqfn)):
                    with open(fqfn, 'wb') as fh:
                        if not isinstance(content, bytes):
                            content = content.encode()
                        fh.write(content)

            # Post File
            url = f'/v2/groups/{api_branch}/{xid}/upload'
            headers = {'Content-Type': 'application/octet-stream'}
            params = {'owner': self._owner, 'updateIfExists': 'true'}
            r = self.submit_file_content('POST', url, content, headers, params, halt_on_error)
            if r is None:
                return None

            if r.status_code == 401:
                # use PUT method if file already exists
                self.log.info('feature=batch, event=401-from-post, action=switch-to-put')
                r = self.submit_file_content('PUT', url, content, headers, params, halt_on_error)
                if r is None:
                    return None

            if not r.ok:
                status = False
                handle_error(
                    code=585,
                    message_values=[r.status_code, r.text],
                    raise_error=halt_on_error,
                )
            elif self.debug and self.enable_saved_file and xid not in self.saved_xids:
                # save xid "if" successfully uploaded and not already saved
                self.saved_xids = xid

            self.log.info(
                f'feature=batch, event=file-upload, status={r.status_code}, '
                f'xid={xid}, remaining={len(file_data)}'
            )
            upload_status.append({'uploaded': status, 'xid': xid})

        return upload_status

    def submit_file_content(
        self,
        method: str,
        url: str,
        data: bytes | str,
        headers: dict,
        params: dict,
        halt_on_error: bool = True,
    ) -> Response | None:
        """Submit File Content for Documents and Reports to ThreatConnect API.

        Args:
            method: The HTTP method for the request (POST, PUT).
            url: The URL for the request.
            data: The body (data) for the request.
            headers: The headers for the request.
            params: The query string parameters for the request.
            halt_on_error: If True any exception will raise an error.

        Returns:
            Response: The response from the request.
        """
        r = None
        try:
            r = self.session_tc.request(method, url, data=data, headers=headers, params=params)
        except Exception as e:
            handle_error(code=580, message_values=[e], raise_error=halt_on_error)
        return r

    def submit_job(self, halt_on_error: bool = True) -> int | None:
        """Submit Batch request to ThreatConnect API.

        Args:
            halt_on_error: If True any exception will raise an error.

        Returns:
            int: The batch id from the API response.
        """
        # check global setting for override
        if self.halt_on_batch_error is not None:
            halt_on_error = self.halt_on_batch_error

        try:
            r = self.session_tc.post('/v2/batch', json=self.settings)
        except Exception as e:
            handle_error(code=10505, message_values=[e], raise_error=halt_on_error)
            return None

        if not r.ok or 'application/json' not in r.headers.get('content-type', ''):
            handle_error(
                code=10510,
                message_values=[r.status_code, r.text],
                raise_error=halt_on_error,
            )
        data = r.json()
        if data.get('status') != 'Success':
            handle_error(
                code=10510,
                message_values=[r.status_code, r.text],
                raise_error=halt_on_error,
            )
        self.log.debug(f'feature=batch, event=submit-job, status={data}')
        return data.get('data', {}).get('batchId')

    def submit_thread(
        self,
        name: str,
        target: Callable,
        args: tuple | None = None,
        kwargs: dict | None = None,
    ):
        """Start a submit thread.

        Args:
            name: The name of the thread.
            target: The method to call for the thread.
            args: The args to pass to the target method.
            kwargs: Additional args.
        """
        self.log.info(f'feature=batch, event=submit-thread, name={name}')
        args = args or ()
        t = None
        try:
            t = threading.Thread(name=name, target=target, args=args, kwargs=kwargs, daemon=True)
            t.start()
        except Exception:
            self.log.trace(traceback.format_exc())
        return t

    def write_error_json(self, errors: list):
        """Write the errors to a JSON file for debugging purposes.

        Args:
            errors: A list of errors to write out.
        """
        if self.debug:
            if not errors:
                errors = []
            # get timestamp as a string without decimal place and consistent length
            timestamp = str(int(time.time() * 10000000))
            error_json_file = os.path.join(self.debug_path_batch, f'errors-{timestamp}.json.gz')
            with gzip.open(error_json_file, mode='wt', encoding='utf-8') as fh:
                json.dump(errors, fh)

    def write_batch_json(self, content: dict):
        """Write batch json data to a file."""
        if self.debug and content:
            # get timestamp as a string without decimal place and consistent length
            timestamp = str(int(time.time() * 10000000))
            batch_json_file = os.path.join(self.debug_path_batch, f'batch-{timestamp}.json.gz')
            with gzip.open(batch_json_file, mode='wt', encoding='utf-8') as fh:
                json.dump(content, fh)

    @property
    def group_len(self) -> int:
        """Return the number of current groups."""
        return len(self.groups) + len(self.groups_shelf)

    @property
    def indicator_len(self) -> int:
        """Return the number of current indicators."""
        return len(self.indicators) + len(self.indicators_shelf)

    def __len__(self) -> int:
        """Return the number of groups and indicators."""
        return self.group_len + self.indicator_len

    def __str__(self) -> str:  # pragma: no cover
        """Return string representation of batch."""
        groups = []
        for group_data in self.groups.values():
            if isinstance(group_data, dict):
                groups.append(group_data)
            else:
                groups.append(group_data.data)
        for group_data in self.groups_shelf.values():
            if isinstance(group_data, dict):
                groups.append(group_data)
            else:
                groups.append(group_data.data)

        indicators = []
        for indicator_data in self.indicators.values():
            if isinstance(indicator_data, dict):
                indicators.append(indicator_data)
            else:
                indicators.append(indicator_data.data)
        for indicator_data in self.indicators_shelf.values():
            if isinstance(indicator_data, dict):
                indicators.append(indicator_data)
            else:
                indicators.append(indicator_data.data)

        data = {'group': groups, 'indicators': indicators}
        return json.dumps(data, indent=4, sort_keys=True)



"""TcEx Framework Module"""
# pylint: disable=no-self-argument,wrong-import-position
# standard library
from typing import Any

# third-party
from pydantic import BaseModel, Field, validator

# first-party
from tcex.input.field_type import EditChoice


class AdvancedRequestModel(BaseModel):
    """Advanced Settings Model

    * why was input included -> feature (what feature?), runtime_level
    * where is input defined -> default (core), install.json

    Feature: advancedRequest

    Supported for the following runtimeLevel:
    * Playbook
    """

    tc_adv_req_body: Any | None = Field(
        None,
        description='The HTTP body for the request.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_exclude_null_params: bool = Field(
        False,
        description='Flag to exclude any null query parameters.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_fail_on_error: bool = Field(
        False,
        description='Flag to force fail on any error.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_headers: list[dict] | None = Field(
        None,
        description='The HTTP headers for the request.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_http_method: EditChoice = Field(
        None,
        description='The HTTP method for the request.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_params: list[dict] | None = Field(
        None,
        description='The HTTP query params for the request.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_path: str | None = Field(
        None,
        description='The API path for the request.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )
    tc_adv_req_urlencode_body: bool = Field(
        False,
        description='Flag to set URL encoding for the request body.',
        inclusion_reason='feature (advancedRequest)',
        requires_definition=True,
    )

    @validator('tc_adv_req_headers', 'tc_adv_req_params', always=True, pre=True)
    def _always_array(cls, value: list | str | None) -> list:
        """Return array value for headers and params."""
        match value:
            case list():
                return value
            case None:
                return []
            case _:
                return [value]



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class TaskFilter(FilterABC):
    """Filter Object for Tasks"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.TASKS.value

    def assigned_to_user_or_group(self, operator: Enum, assigned_to_user_or_group: list | str):
        """Filter Assigned To User or Group based on **assignedToUserOrGroup** keyword.

        Args:
            operator: The operator enum for the filter.
            assigned_to_user_or_group: A value of User, Group, or None depending on the assignee.
        """
        if isinstance(assigned_to_user_or_group, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter(
            'assignedToUserOrGroup', operator, assigned_to_user_or_group, TqlType.STRING
        )

    def assignee_name(self, operator: Enum, assignee_name: list | str):
        """Filter Assignee Name based on **assigneeName** keyword.

        Args:
            operator: The operator enum for the filter.
            assignee_name: The user or group name assigned to the Task.
        """
        if isinstance(assignee_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('assigneeName', operator, assignee_name, TqlType.STRING)

    def automated(self, operator: Enum, automated: bool):
        """Filter Automated based on **automated** keyword.

        Args:
            operator: The operator enum for the filter.
            automated: A flag indicating whether or not the task is automated.
        """
        self._tql.add_filter('automated', operator, automated, TqlType.BOOLEAN)

    def case_id(self, operator: Enum, case_id: int | list):
        """Filter Case ID based on **caseId** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id: The ID of the case this Task is associated with.
        """
        if isinstance(case_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('caseId', operator, case_id, TqlType.INTEGER)

    def case_id_as_string(self, operator: Enum, case_id_as_string: list | str):
        """Filter CaseID As String based on **caseIdAsString** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id_as_string: The ID of the case as a String.
        """
        if isinstance(case_id_as_string, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('caseIdAsString', operator, case_id_as_string, TqlType.STRING)

    def case_severity(self, operator: Enum, case_severity: list | str):
        """Filter Case Severity based on **caseSeverity** keyword.

        Args:
            operator: The operator enum for the filter.
            case_severity: The severity of the case associated with the task.
        """
        if isinstance(case_severity, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('caseSeverity', operator, case_severity, TqlType.STRING)

    def completed_by(self, operator: Enum, completed_by: list | str):
        """Filter Completed By based on **completedBy** keyword.

        Args:
            operator: The operator enum for the filter.
            completed_by: The account login of the user who completed the task.
        """
        if isinstance(completed_by, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('completedBy', operator, completed_by, TqlType.STRING)

    def completed_date(self, operator: Enum, completed_date: Arrow | datetime | int | str):
        """Filter Completed Date based on **completedDate** keyword.

        Args:
            operator: The operator enum for the filter.
            completed_date: The completion date for the task.
        """
        completed_date = self.util.any_to_datetime(completed_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('completedDate', operator, completed_date, TqlType.STRING)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of the task.
        """
        if isinstance(description, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def due_date(self, operator: Enum, due_date: Arrow | datetime | int | str):
        """Filter Due Date based on **dueDate** keyword.

        Args:
            operator: The operator enum for the filter.
            due_date: The due date for the task.
        """
        due_date = self.util.any_to_datetime(due_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dueDate', operator, due_date, TqlType.STRING)

    @property
    def has_artifact(self):
        """Return **ArtifactFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter

        artifacts = ArtifactFilter(Tql())
        self._tql.add_filter('hasArtifact', TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)
        return artifacts

    @property
    def has_case(self):
        """Return **CaseFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.cases.case_filter import CaseFilter

        cases = CaseFilter(Tql())
        self._tql.add_filter('hasCase', TqlOperator.EQ, cases, TqlType.SUB_QUERY)
        return cases

    @property
    def has_note(self):
        """Return **NoteFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.notes.note_filter import NoteFilter

        notes = NoteFilter(Tql())
        self._tql.add_filter('hasNote', TqlOperator.EQ, notes, TqlType.SUB_QUERY)
        return notes

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the task.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def missing_artifact_count(self, operator: Enum, missing_artifact_count: int | list):
        """Filter Missing Artifact Count based on **missingArtifactCount** keyword.

        Args:
            operator: The operator enum for the filter.
            missing_artifact_count: The number of required artifacts that are missing.
        """
        if isinstance(missing_artifact_count, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter(
            'missingArtifactCount', operator, missing_artifact_count, TqlType.INTEGER
        )

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of the task.
        """
        if isinstance(name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the case.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name for the case.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def required(self, operator: Enum, required: bool):
        """Filter Required based on **required** keyword.

        Args:
            operator: The operator enum for the filter.
            required: Flag indicating whether or not the task is required.
        """
        self._tql.add_filter('required', operator, required, TqlType.BOOLEAN)

    def status(self, operator: Enum, status: list | str):
        """Filter Status based on **status** keyword.

        Args:
            operator: The operator enum for the filter.
            status: The status of the task.
        """
        if isinstance(status, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('status', operator, status, TqlType.STRING)

    def target_id(self, operator: Enum, target_id: int | list):
        """Filter Assignee based on **targetId** keyword.

        Args:
            operator: The operator enum for the filter.
            target_id: The assigned user or group ID for the task.
        """
        if isinstance(target_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('targetId', operator, target_id, TqlType.INTEGER)

    def target_type(self, operator: Enum, target_type: list | str):
        """Filter Target Type based on **targetType** keyword.

        Args:
            operator: The operator enum for the filter.
            target_type: The target type for this task (either User or Group).
        """
        if isinstance(target_type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('targetType', operator, target_type, TqlType.STRING)

    def workflow_phase(self, operator: Enum, workflow_phase: int | list):
        """Filter Workflow Phase based on **workflowPhase** keyword.

        Args:
            operator: The operator enum for the filter.
            workflow_phase: The workflow phase of the task.
        """
        if isinstance(workflow_phase, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('workflowPhase', operator, workflow_phase, TqlType.INTEGER)

    def workflow_step(self, operator: Enum, workflow_step: int | list):
        """Filter Workflow Step based on **workflowStep** keyword.

        Args:
            operator: The operator enum for the filter.
            workflow_step: The workflow step of the task.
        """
        if isinstance(workflow_step, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('workflowStep', operator, workflow_step, TqlType.INTEGER)

    def xid(self, operator: Enum, xid: list | str):
        """Filter XID based on **xid** keyword.

        Args:
            operator: The operator enum for the filter.
            xid: The XID of the task.
        """
        if isinstance(xid, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('xid', operator, xid, TqlType.STRING)



"""TcEx Framework Module"""
# standard library
from collections.abc import Iterator

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.user_groups.user_group_filter import UserGroupFilter
from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel, UserGroupsModel


class UserGroup(ObjectABC):
    """UserGroups Object."""

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: UserGroupModel = UserGroupModel(**kwargs)
        self._nested_field_name = 'userGroups'
        self._nested_filter = 'has_user_group'
        self.type_ = 'User Group'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.USER_GROUPS.value

    @property
    def model(self) -> UserGroupModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | UserGroupModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')


class UserGroups(ObjectCollectionABC):
    """UserGroups Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = UserGroupsModel(**kwargs)
        self.type_ = 'user_groups'

    def __iter__(self) -> Iterator[UserGroup]:
        """Return CM objects."""
        return self.iterate(base_class=UserGroup)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.USER_GROUPS.value

    @property
    def filter(self) -> UserGroupFilter:
        """Return the type specific filter object."""
        return UserGroupFilter(self.tql)



"""TcEx Framework Module"""
# third-party
from pydantic import BaseModel, Field


class LoggingModel(BaseModel):
    """Logging Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * WebhookTriggerService
    * TriggerService
    """

    # target 50Mb in total log size (1x10Mb + 25x~1.6Mb = ~50Mb)
    tc_log_backup_count: int = Field(
        25,
        description='The maximum number of log files to retain for an App.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_log_file: str = Field(
        'app.log',
        description='The default name of the App\'s log file.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    # job Apps have to collect tc_log_level manually
    tc_log_level: str = Field(
        'info',
        description='The logging level for the App.',
        inclusion_reason='runtimeLevel',
        # requires_definition=True, # this is true for only Organization runtimeLevel
    )
    tc_log_max_bytes: int = Field(
        10_485_760,
        description='The maximum size of the App log file before rotation.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_log_to_api: bool = Field(
        False,
        description='Flag to enable API logging for the App.',
        inclusion_reason='runtimeLevel',
    )



"""TcEx Framework Module"""
# standard library
from abc import ABC, abstractmethod
from typing import Any


class KeyValueABC(ABC):
    """Abstract base class for all KeyValue clients."""

    @abstractmethod
    def create(self, context: str, key: str, value: Any) -> int:
        """Create key/value pair in remote KV store.

        Args:
            context: A specific context for the create.
            key: The key to create in KV store.
            value: The value to store in KV store.

        Returns:
            (string): The response from the KV store provider.
        """

    @abstractmethod
    def read(self, context: str, key: str) -> Any:
        """Read data from KV store for the provided key.

        Args:
            context: A specific context for the create.
            key: The key to read in KV store.

        Returns:
            (any): The response data from the  KV store provider.
        """
